"""Private module that determines how data is encoded and serialized, to be able to send it over a wire, or save to disk"""

import base64
import io
import json
import numbers
import uuid
import struct

import numpy as np
import pyarrow as pa
import vaex

registry = {}


def register(name):
    def wrapper(cls):
        assert name not in registry
        registry[name] = cls
        return cls
    return wrapper


@register("json")  # this will pass though data as is
class vaex_json_encoding:
    @classmethod
    def encode(cls, encoding, result):
        return result

    @classmethod
    def decode(cls, encoding, result_encoded):
        return result_encoded


@register("vaex-task-result")
class vaex_task_result_encoding:
    @classmethod
    def encode(cls, encoding, result):
        return encoding.encode('vaex-evaluate-result', result)

    @classmethod
    def decode(cls, encoding, result_encoded):
        return encoding.decode('vaex-evaluate-result', result_encoded)


@register("vaex-rmi-result")
class vaex_rmi_result_encoding:
    @classmethod
    def encode(cls, encoding, result):
        return encoding.encode('json', result)

    @classmethod
    def decode(cls, encoding, result_encoded):
        return encoding.decode('json', result_encoded)


@register("vaex-evaluate-result")
class vaex_evaluate_results_encoding:
    @classmethod
    def encode(cls, encoding, result):
        if isinstance(result, (list, tuple)):
            return [cls.encode(encoding, k) for k in result]
        else:
            if isinstance(result, np.ndarray):
                return {'type': 'ndarray', 'data': encoding.encode('ndarray', result)}
            elif isinstance(result, vaex.array_types.supported_arrow_array_types):
                return {'type': 'arrow-array', 'data': encoding.encode('arrow-array', result)}
            elif isinstance(result, numbers.Number):
                try:
                    result = result.item()  # for numpy scalars
                except:  # noqa
                    pass
                return {'type': 'json', 'data': result}
            else:
                raise ValueError('Cannot encode: %r' % result)

    @classmethod
    def decode(cls, encoding, result_encoded):
        if isinstance(result_encoded, (list, tuple)):
            return [cls.decode(encoding, k) for k in result_encoded]
        else:
            return encoding.decode(result_encoded['type'], result_encoded['data'])



@register("arrow-array")
class arrow_array_encoding:
    @classmethod
    def encode(cls, encoding, array):
        blob = pa.serialize(array).to_buffer()
        return {'arrow-serialized-blob': encoding.add_blob(blob)}

    @classmethod
    def decode(cls, encoding, result_encoded):
        blob = encoding.get_blob(result_encoded['arrow-serialized-blob'])
        return pa.deserialize(blob)


@register("ndarray")
class ndarray_encoding:
    @classmethod
    def encode(cls, encoding, array):
        # if array.dtype.kind == 'O':
        #     raise ValueError('Numpy arrays with objects cannot be serialized: %r' % array)
        mask = None
        dtype = array.dtype
        if np.ma.isMaskedArray(array):
            values = array.data
            mask = array.mask
        else:
            values = array
        if values.dtype.kind in 'mM':
            values = values.view(np.uint64)
        if values.dtype.kind == 'O':
            data = {
                    'values': values.tolist(),  # rely on json encoding
                    'shape': array.shape,
                    'dtype': encoding.encode('dtype', dtype)
            }
        else:
            data = {
                    'values': encoding.add_blob(values),
                    'shape': array.shape,
                    'dtype': encoding.encode('dtype', dtype)
            }
        if mask is not None:
            data['mask'] = encoding.add_blob(mask)
        return data

    @classmethod
    def decode(cls, encoding, result_encoded):
        if isinstance(result_encoded, (list, tuple)):
            return [cls.decode(encoding, k) for k in result_encoded]
        else:
            dtype = encoding.decode('dtype', result_encoded['dtype'])
            shape = result_encoded['shape']
            if dtype.kind == 'O':
                data = result_encoded['values']
                array = np.array(data, dtype=dtype)
            else:
                data = encoding.get_blob(result_encoded['values'])
                array = np.frombuffer(data, dtype=dtype).reshape(shape)
            if 'mask' in result_encoded:
                mask_data = encoding.get_blob(result_encoded['mask'])
                mask_array = np.frombuffer(mask_data, dtype=np.bool_).reshape(shape)
                array = np.ma.array(array, mask=mask_array)
            return array


@register("dtype")
class dtype_encoding:
    @staticmethod
    def encode(encoding, dtype):
        if type(dtype) == type:
            dtype = dtype().dtype
        return str(dtype)

    @staticmethod
    def decode(encoding, type_spec):
        if type_spec == 'string':
            return pa.string()
        if type_spec == 'large_string':
            return pa.large_string()
        # TODO: find a proper way to support all arrow types
        if type_spec == 'timestamp[ms]':
            return pa.timestamp('ms')
        else:
            return np.dtype(type_spec)


@register("binner")
class binner_encoding:
    @staticmethod
    def encode(encoding, binner):
        name = type(binner).__name__
        if name.startswith('BinnerOrdinal_'):
            datatype = name[len('BinnerOrdinal_'):]
            if datatype.endswith("_non_native"):
                datatype = datatype[:-len('64_non_native')]
                datatype = encoding.encode('dtype', np.dtype(datatype).newbyteorder())
            return {'type': 'ordinal', 'expression': binner.expression, 'datatype': datatype, 'count': binner.ordinal_count, 'minimum': binner.min_value}
        elif name.startswith('BinnerScalar_'):
            datatype = name[len('BinnerScalar_'):]
            if datatype.endswith("_non_native"):
                datatype = datatype[:-len('64_non_native')]
                datatype = encoding.encode('dtype', np.dtype(datatype).newbyteorder())
            return {'type': 'scalar', 'expression': binner.expression, 'datatype': datatype, 'count': binner.bins, 'minimum': binner.vmin, 'maximum': binner.vmax}
        else:
            raise ValueError('Cannot serialize: %r' % binner)

    @staticmethod
    def decode(encoding, binner_spec):
        type = binner_spec['type']
        dtype = encoding.decode('dtype', binner_spec['datatype'])
        if type == 'ordinal':
            cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerOrdinal_", dtype)
            return cls(binner_spec['expression'], binner_spec['count'], binner_spec['minimum'])
        elif type == 'scalar':
            cls = vaex.utils.find_type_from_dtype(vaex.superagg, "BinnerScalar_", dtype)
            return cls(binner_spec['expression'], binner_spec['minimum'], binner_spec['maximum'], binner_spec['count'])
        else:
            raise ValueError('Cannot deserialize: %r' % binner_spec)


@register("grid")
class grid_encoding:
    @staticmethod
    def encode(encoding, grid):
        return encoding.encode_list('binner', grid.binners)

    @staticmethod
    def decode(encoding, grid_spec):
        return vaex.superagg.Grid(encoding.decode_list('binner', grid_spec))


class Encoding:
    def __init__(self, next=None):
        self.registry = {**registry}
        self.blobs = {}

    def encode(self, typename, value):
        encoded = self.registry[typename].encode(self, value)
        return encoded

    def encode_list(self, typename, values):
        encoded = [self.registry[typename].encode(self, k) for k in values]
        return encoded

    def encode_list2(self, typename, values):
        encoded = [self.encode_list(typename, k) for k in values]
        return encoded

    def encode_dict(self, typename, values):
        encoded = {key: self.registry[typename].encode(self, value) for key, value in values.items()}
        return encoded

    def decode(self, typename, value, **kwargs):
        decoded = self.registry[typename].decode(self, value, **kwargs)
        return decoded

    def decode_list(self, typename, values, **kwargs):
        decoded = [self.registry[typename].decode(self, k, **kwargs) for k in values]
        return decoded

    def decode_list2(self, typename, values, **kwargs):
        decoded = [self.decode_list(typename, k, **kwargs) for k in values]
        return decoded

    def decode_dict(self, typename, values, **kwargs):
        decoded = {key: self.registry[typename].decode(self, value, **kwargs) for key, value in values.items()}
        return decoded

    def add_blob(self, buffer):
        blob_id = str(uuid.uuid4())
        self.blobs[blob_id] = memoryview(buffer).tobytes()
        return f'blob:{blob_id}'

    def get_blob(self, blob_ref):
        assert blob_ref.startswith('blob:')
        blob_id = blob_ref[5:]
        return self.blobs[blob_id]


class inline:
    @staticmethod
    def serialize(data, encoding):
        import base64
        blobs = {key: base64.b64encode(value).decode('ascii') for key, value in encoding.blobs.items()}
        return json.dumps({'data': data, 'blobs': blobs})

    @staticmethod
    def deserialize(data, encoding):
        data = json.loads(data)
        encoding.blobs = {key: base64.b64decode(value.encode('ascii')) for key, value in data['blobs'].items()}
        return data['data']


def _pack_blobs(*blobs):
    count = len(blobs)
    lenghts = [len(blob) for blob in blobs]
    stream = io.BytesIO()
    # header: <number of blobs>,<offset 0>, ... <offset N-1> with 8 byte unsigned ints
    header_length = 8 * (2 + count)
    offsets = (np.cumsum([0] + lenghts) + header_length).tolist()
    stream.write(struct.pack(f'{count+2}q', count, *offsets))
    for blob in blobs:
        stream.write(blob)
    bytes = stream.getvalue()
    assert offsets[-1] == len(bytes)
    return bytes


def _unpack_blobs(bytes):
    stream = io.BytesIO(bytes)

    count, = struct.unpack('q', stream.read(8))
    offsets = struct.unpack(f'{count+1}q', stream.read(8 * (count + 1)))
    assert offsets[-1] == len(bytes)
    blobs = []
    for i1, i2 in zip(offsets[:-1], offsets[1:]):
        blobs.append(bytes[i1:i2])
    return blobs


class binary:
    @staticmethod
    def serialize(data, encoding):
        blob_refs = list(encoding.blobs.keys())
        blobs = [encoding.blobs[k] for k in blob_refs]
        json_blob = json.dumps({'data': data, 'blob_refs': blob_refs})
        return _pack_blobs(json_blob.encode('utf8'), *blobs)

    @staticmethod
    def deserialize(data, encoding):
        json_data, *blobs = _unpack_blobs(data)
        json_data = json_data.decode('utf8')
        json_data = json.loads(json_data)
        data = json_data['data']
        encoding.blobs = {key: blob for key, blob in zip(json_data['blob_refs'], blobs)}
        return data


serialize = binary.serialize
deserialize = binary.deserialize



__author__ = 'maartenbreddels'
import collections
import logging

import pyarrow as pa
import pyarrow.dataset

import vaex.dataset
from ..itertools import buffer
from vaex.multithreading import get_main_io_pool


logger = logging.getLogger("vaex.multithreading")


class DatasetArrow(vaex.dataset.Dataset):
    def __init__(self, ds, max_rows_read=1024**2*10):
        super().__init__()
        self.max_rows_read = max_rows_read
        self._arrow_ds = ds
        row_count = 0
        for fragment in self._arrow_ds.get_fragments():
            if hasattr(fragment, "ensure_complete_metadata"):
                fragment.ensure_complete_metadata()
            for rg in fragment.row_groups:
                row_count += rg.num_rows
        self._row_count = row_count
        self._columns = {name: vaex.dataset.ColumnProxy(self, name, type) for name, type in
                          zip(self._arrow_ds.schema.names, self._arrow_ds.schema.types)}
        self._ids = {}

    def hashed(self):
        raise NotImplementedError

    def slice(self, start, end):
        # TODO: we can be smarter here, and trim off some fragments
        if start == 0 and end == self.row_count:
            return self
        return vaex.dataset.DatasetSliced(self, start=start, end=end)

    def is_masked(self, column):
        return False

    def shape(self, column):
        return tuple()

    def __getitem__(self, item):
        if isinstance(item, slice):
            assert item.step in [1, None]
            return vaex.dataset.DatasetSliced(self, item.start or 0, item.stop or self.row_count)
        return self._columns[item]

    def close(self):
        # no need to close it, it seem
        pass

    def _chunk_producer(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        import pyarrow.parquet
        pool = get_main_io_pool()
        offset = 0
        for fragment_large in self._arrow_ds.get_fragments():
            fragment_large_rows = sum([rg.num_rows for rg in fragment_large.row_groups])
            fragments = [fragment_large]
            # when do we want to split up? File size? max chunk size?
            if fragment_large_rows > self.max_rows_read:
                fragments = fragment_large.split_by_row_group()
            for fragment in fragments:
                rows = sum([rg.num_rows for rg in fragment.row_groups])
                chunk_start = offset
                chunk_end = offset + rows

                length = chunk_end - chunk_start  # default length

                if start >= chunk_end:  # we didn't find the beginning yet
                    offset += length
                    continue
                if end <= chunk_start:  # we are past the end
                    # assert False
                    break
                def reader(fragment=fragment):
                    table = fragment.to_table(columns=columns, use_threads=False)
                    chunks = dict(zip(table.column_names, table.columns))
                    return chunks
                assert length > 0
                if start > chunk_start:
                    # this means we have to cut off a piece of the beginning
                    if end < chunk_end:
                        # AND the end
                        length = end - chunk_start  # without the start cut off
                        length -= start - chunk_start  # correcting for the start cut off
                        assert length > 0
                        def slicer(chunk_start=chunk_start, reader=reader, length=length):
                            chunks = reader()
                            chunks = {name: ar.slice(start - chunk_start, length) for name, ar in chunks.items()}
                            for name, ar in chunks.items():
                                assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'
                            return chunks
                        reader = slicer
                    else:
                        length -= start - chunk_start  # correcting for the start cut off
                        assert length > 0
                        def slicer(chunk_start=chunk_start, reader=reader, length=length):
                            chunks = reader()
                            chunks = {name: ar.slice(start - chunk_start) for name, ar in chunks.items()}
                            for name, ar in chunks.items():
                                assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'
                            return chunks
                        reader = slicer
                else:
                    if end < chunk_end:
                        # we only need to cut off a piece of the end
                        length = end - chunk_start
                        assert length > 0
                        def slicer(chunk_start=chunk_start, reader=reader, length=length):
                            chunks = reader()
                            chunks = {name: ar.slice(0, length) for name, ar in chunks.items()}
                            for name, ar in chunks.items():
                                assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'
                            return chunks
                        reader = slicer
                offset += rows
                yield pool.submit(reader)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        chunk_size = chunk_size or 1024*1024
        i1 = 0
        chunks_ready_list = []
        i1 = i2 = 0

        workers = get_main_io_pool()._max_workers
        for chunks_future in buffer(self._chunk_producer(columns, chunk_size, start=start, end=end or self._row_count), workers+3):
            chunks = chunks_future.result()
            chunks_ready_list.append(chunks)
            total_row_count = sum([len(list(k.values())[0]) for k in chunks_ready_list])
            if total_row_count > chunk_size:
                chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
                i2 += current_row_count
                yield i1, i2, vaex.dataset._concat_chunk_list(chunks_current_list)
                i1 = i2

        while chunks_ready_list:
            chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
            i2 += current_row_count
            yield i1, i2, vaex.dataset._concat_chunk_list(chunks_current_list)
            i1 = i2



def from_table(table):
    columns = dict(zip(table.schema.names, table.columns))
    dataset = vaex.dataset.DatasetArrays(columns)
    return dataset


def open(path, fs_options):
    source = vaex.file.open_for_arrow(path=path, mode='rb', fs_options=fs_options, mmap=True)
    try:
        # first we try if it opens as stream
        reader = pa.ipc.open_stream(source)
    except pa.lib.ArrowInvalid:
        # if not, we open as file
        reader = pa.ipc.open_file(source)
        # for some reason this reader is not iterable
        batches = [reader.get_batch(i) for i in range(reader.num_record_batches)]
    else:
        # if a stream, we're good
        batches = reader  # this reader is iterable
    table = pa.Table.from_batches(batches)
    return from_table(table)


def open_parquet(path, fs_options={}):
    import vaex.file.arrow
    file_system, path = vaex.file.arrow.parse(path, fs_options)
    arrow_ds = pyarrow.dataset.dataset(path, filesystem=file_system)
    return DatasetArrow(arrow_ds)




import asyncio
import sys


def check_ipython():
    IPython = sys.modules.get('IPython')
    if IPython:
        IPython_version = tuple(map(int, IPython.__version__.split('.')))
        if IPython_version < (7, 0, 0):
            raise RuntimeError(f'You are using IPython {IPython.__version__} while we require 7.0.0, please update IPython')

def check_patch_tornado():
    '''If tornado is imported, add the patched asyncio.Future to its tuple of acceptable Futures'''
    if 'tornado' in sys.modules:
        import tornado.concurrent
        if asyncio.Future not in tornado.concurrent.FUTURES:
            tornado.concurrent.FUTURES = tornado.concurrent.FUTURES + (asyncio.Future, )            


def just_run(coro):
    '''Make the coroutine run, even if there is already an event loop running (using nest_asyncio)'''
    try:
        loop = asyncio.get_event_loop()
        had_loop = True
    except RuntimeError:
        had_loop = False
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    try:
        if had_loop:
            check_ipython()
            import nest_asyncio
            nest_asyncio.apply()
            check_patch_tornado()
        return loop.run_until_complete(coro)
    finally:
        if not had_loop:  # remove loop if we did not have one
            asyncio.set_event_loop(None)



"""Implementation all tasks parts for the cpu"""

from functools import reduce
import numpy as np

import vaex
import vaex.encoding
from .utils import as_flat_float, as_flat_array, _issequence, _ensure_list
from . import array_types
from .array_types import filter


_task_part_register = {}


def register(cls):
    assert cls is not None
    _task_part_register[cls.name] = cls
    return cls


def create_part_from_spec(df, spec):
    cls = _task_part_register[spec['task']]
    part_cls = cls.part_class
    del spec['task']
    return part_cls.from_spec(df, spec)


@vaex.encoding.register("task-part-cpu")
class encoder:
    @staticmethod
    def encode(encoding, task):
        return task.encode(encoding)

    @staticmethod
    def decode(encoding, spec, df):
        cls = _task_part_register[spec['task']]
        return cls.decode(encoding, spec, df)


class TaskPart:
    def __init__(self, df, expressions, name, pre_filter):
        self.df = df
        self.expressions = expressions
        self.name = name
        self.pre_filter = pre_filter


@register
class TaskPartSum:
    name = "sum-test"

    def __init__(self, expression):
        self.total = 0
        self.expression = expression

    @property
    def expressions(self):
        return [self.expression]

    def get_result(self):
        return self.total

    def process(self, thread_index, i1, i2, filter_mask, chunk):
        self.total += chunk.sum()

    def reduce(self, others):
        self.total += sum(other.total for other in others)

    @classmethod
    def decode(cls, encoding, spec, df):
        return cls(spec['expression'])


@register
class TaskPartMapReduce(TaskPart):
    name = "map_reduce"

    def __init__(self, df, expressions, map, reduce, converter=lambda x: x, info=False, to_float=False,
                 to_numpy=True, ordered_reduce=False, skip_masked=False, ignore_filter=False, selection=None, pre_filter=False, name="task"):
        TaskPart.__init__(self, df, expressions, name=name, pre_filter=pre_filter)

        self._map = map
        self._reduce = reduce
        self.converter = converter
        self.info = info
        self.ordered_reduce = ordered_reduce
        self.to_float = to_float
        self.to_numpy = to_numpy
        self.skip_masked = skip_masked
        self.ignore_filter = ignore_filter
        if self.pre_filter and self.ignore_filter:
            raise ValueError("Cannot pre filter and also ignore the filter")
        self.selection = selection
        self.values = []

    def process(self, thread_index, i1, i2, filter_mask, *blocks):
        if self.to_numpy:
            blocks = [block if isinstance(block, np.ndarray) else block.to_numpy() for block in blocks]
        if self.to_float:
            blocks = [as_flat_float(block) for block in blocks]
        if self.skip_masked:
            masks = [np.ma.getmaskarray(block) for block in blocks if np.ma.isMaskedArray(block)]
            blocks = [block.data if np.ma.isMaskedArray(block) else block for block in blocks]
            mask = None
            if masks:
                # find all 'rows', where all columns are present (not masked)
                mask = masks[0].copy()
                for other in masks[1:]:
                    mask |= other
                blocks = [block[~mask] for block in blocks]

        if not self.ignore_filter:
            selection = self.selection
            if self.pre_filter:
                if selection:
                    selection_mask = self.df.evaluate_selection_mask(selection, i1=i1, i2=i2, cache=True)
                    blocks = [filter(block, selection_mask) for block in blocks]
            else:
                if selection or self.df.filtered:
                    selection_mask = self.df.evaluate_selection_mask(selection, i1=i1, i2=i2, cache=True, pre_filtered=False)
                    if filter_mask is not None:
                        selection_mask = selection_mask & filter_mask
                    blocks = [filter(block, selection_mask) for block in blocks]
        if self.info:
            self.values.append(self._map(thread_index, i1, i2, *blocks))
        else:
            self.values.append(self._map(*blocks))

    def reduce(self, others):
        # if self.ordered_reduce:
        #     results.sort(key=lambda x: x[0])
        #     results = [k[1] for k in results]
        # return self.converter(reduce(self._reduce, results))
        values = self.values
        for other in others:
            values.extend(other.values)
        if values:
            self.values = reduce(self._reduce, values)

    def get_result(self):
        return self.values

    @classmethod
    def decode(cls, encoding, spec, df):
        spec = spec.copy()
        del spec['task']
        return cls(df, **spec)


@register
class TaskPartStatistic:
    name = "legacy_statistic"

    def __init__(self, df, shape, expressions, dtype, selections, op, weights, minima, maxima, edges, selection_waslist):
        self.df = df
        self.shape = shape
        self.dtype = dtype
        self.expressions = expressions
        self.op = op
        self.selections = selections
        self.fields = op.fields(weights)
        self.shape_total = (len(self.selections), ) + self.shape + (self.fields,)
        self.grid = np.zeros(self.shape_total, dtype=self.dtype)
        self.op.init(self.grid)
        self.minima = minima
        self.maxima = maxima
        self.edges = edges
        self.selection_waslist = selection_waslist

    def process(self, thread_index, i1, i2, filter_mask, *blocks):
        class Info(object):
            pass
        info = Info()
        info.i1 = i1
        info.i2 = i2
        info.first = i1 == 0
        info.last = i2 == self.df.length_unfiltered()
        info.size = i2 - i1

        masks = [np.ma.getmaskarray(block) for block in blocks if np.ma.isMaskedArray(block)]
        blocks = [block.data if np.ma.isMaskedArray(block) else block for block in blocks]
        blocks = [np.asarray(k) for k in blocks]
        mask = None

        # blocks = [as_flat_float(block) for block in blocks]
        if len(blocks) != 0:
            dtype = np.find_common_type([block.dtype for block in blocks], [])
            if dtype.str in ">f8 <f8 =f8":
                statistic_function = vaex.vaexfast.statisticNd_f8
            elif dtype.str in ">f4 <f4 =f4":
                statistic_function = vaex.vaexfast.statisticNd_f4
            elif dtype.str in ">i8 <i8 =i8":
                dtype = np.dtype(np.float64)
                statistic_function = vaex.vaexfast.statisticNd_f8
            else:
                dtype = np.dtype(np.float32)
                statistic_function = vaex.vaexfast.statisticNd_f4
            # print(dtype, statistic_function, histogram2d)

        if masks:
            mask = masks[0].copy()
            for other in masks[1:]:
                mask |= other
            blocks = [block[~mask] for block in blocks]

        this_thread_grid = self.grid
        for i, selection in enumerate(self.selections):
            if selection:
                selection_mask = self.df.evaluate_selection_mask(selection, i1=i1, i2=i2, cache=True)  # TODO
                if selection_mask is None:
                    raise ValueError("performing operation on selection while no selection present")
                if mask is not None:
                    selection_mask = selection_mask[~mask]
                selection_blocks = [block[selection_mask] for block in blocks]
            else:
                selection_blocks = [block for block in blocks]
            little_endians = len([k for k in selection_blocks if k.dtype != str and k.dtype.byteorder in ["<", "="]])
            if not ((len(selection_blocks) == little_endians) or little_endians == 0):
                def _to_native(ar):
                    if ar.dtype == str:
                        return ar  # string are always fine
                    if ar.dtype.byteorder not in ["<", "="]:
                        dtype = ar.dtype.newbyteorder()
                        return ar.astype(dtype)
                    else:
                        return ar

                selection_blocks = [_to_native(k) for k in selection_blocks]
            # subblock_weight = None
            subblock_weights = selection_blocks[len(self.expressions):]
            selection_blocks = list(selection_blocks[:len(self.expressions)])
            if len(selection_blocks) == 0 and subblock_weights == []:
                if self.op == vaex.tasks.OP_ADD1:  # special case for counting '*' (i.e. the number of rows)
                    if selection or self.df.filtered:
                        this_thread_grid[i][0] += np.sum(selection_mask)
                    else:
                        this_thread_grid[i][0] += i2 - i1
                else:
                    raise ValueError("Nothing to compute for OP %s" % self.op.code)
            # special case for counting string values etc
            elif len(selection_blocks) == 0 and len(subblock_weights) == 1 and self.op in [vaex.tasks.OP_COUNT]\
                    and (subblock_weights[0].dtype == str or subblock_weights[0].dtype.kind not in 'biuf'):
                weight = subblock_weights[0]
                mask = None
                if weight.dtype != str:
                    if weight.dtype.kind == 'O':
                        mask = vaex.strings.StringArray(weight).mask()
                else:
                    mask = weight.get_mask()
                if selection or self.df.filtered:
                    if mask is not None:
                        this_thread_grid[i][0] += np.sum(~mask)
                    else:
                        this_thread_grid[i][0] += np.sum(selection_mask)
                else:
                    if mask is not None:
                        this_thread_grid[i][0] += len(mask) - mask.sum()
                    else:
                        this_thread_grid[i][0] += len(weight)
            else:
                selection_blocks = [as_flat_array(block, dtype) for block in selection_blocks]
                subblock_weights = [as_flat_array(block, dtype) for block in subblock_weights]
                statistic_function(selection_blocks, subblock_weights, this_thread_grid[i], self.minima, self.maxima, self.op.code, self.edges)
        return i2 - i1
        # return map(self._map, blocks)#[self.map(block) for block in blocks]

    def reduce(self, others):
        grids = [self.grid] + [k.grid for k in others]
        self.grid = self.op.reduce(np.array(grids))
        # If selection was a string, we just return the single selection

    def get_result(self):
        return self.grid if self.selection_waslist else self.grid[0]

    @classmethod
    def decode(cls, encoding, spec, df):
        spec = spec.copy()
        del spec['task']
        spec['op'] = encoding.decode('_op', spec['op'])
        spec['dtype'] = encoding.decode('dtype', spec['dtype'])
        return cls(df, **spec)


@register
class TaskPartAggregations:
    name = "aggregations"

    def __init__(self, df, grid, aggregation_descriptions, dtypes, initial_values=None):
        self.df = df
        self.has_values = False
        self.dtypes = dtypes
        # self.expressions_all = expressions
        self.expressions = [binner.expression for binner in grid.binners]
        # TODO: selection and edges in descriptor?
        self.aggregation_descriptions = aggregation_descriptions
        for aggregator_descriptor in self.aggregation_descriptions:
            self.expressions.extend(aggregator_descriptor.expressions)
        # self.expressions = list(set(expressions))
        self.grid = vaex.superagg.Grid([binner.copy() for binner in grid.binners])

        def create_aggregator(aggregator_descriptor, selections, initial_values):
            # for each selection, we have a separate aggregator, sharing the grid and binners
            for i, selection in enumerate(selections):
                agg = aggregator_descriptor._create_operation(self.df, self.grid)
                if initial_values is not None:
                    print(np.asarray(agg))
                    print(initial_values[i])
                    # np.asarray(agg)[:] = initial_values[i]
                    np.copyto(np.asarray(agg), initial_values[i])
                yield agg

        self.aggregations = []
        for i, aggregator_descriptor in enumerate(self.aggregation_descriptions):
            selection = aggregator_descriptor.selection
            selection_waslist = _issequence(selection)
            selections = _ensure_list(selection)
            initial_values_i = initial_values[i] if initial_values else None
            self.aggregations.append((aggregator_descriptor, selections, list(create_aggregator(aggregator_descriptor, selections, initial_values_i)), selection_waslist))

    def process(self, thread_index, i1, i2, filter_mask, *blocks):
        # self.check()
        grid = self.grid

        def check_array(x, dtype):
            if vaex.array_types.is_string_type(dtype):
                x = vaex.column._to_string_sequence(x)
            else:
                x = vaex.utils.as_contiguous(x)
                if x.dtype.kind in "mM":
                    # we pass datetime as int
                    x = x.view('uint64')
            return x

        N = i2 - i1
        if filter_mask is not None:
            N = filter_mask.astype(np.uint8).sum()
        blocks = [array_types.to_numpy(block, strict=False) for block in blocks]
        for block in blocks:
            assert len(block) == N, f'Oops, got a block of length {len(block)} while it is expected to be of length {N} (at {i1}-{i2}, filter={filter_mask is not None})'
        block_map = {expr: block for expr, block in zip(self.expressions, blocks)}
        # we need to make sure we keep some objects alive, since the c++ side does not incref
        # on set_data and set_data_mask
        references = []
        for binner in grid.binners:
            block = block_map[binner.expression]
            dtype = self.dtypes[binner.expression]
            block = check_array(block, dtype)
            if np.ma.isMaskedArray(block):
                block, mask = block.data, np.ma.getmaskarray(block)
                binner.set_data(block)
                binner.set_data_mask(mask)
                references.extend([block, mask])
            else:
                binner.set_data(block)
                references.extend([block])
        all_aggregators = []
        for agg_desc, selections, aggregation2d, selection_waslist in self.aggregations:
            for selection_index, selection in enumerate(selections):
                agg = aggregation2d[selection_index]
                all_aggregators.append(agg)
                selection_mask = None
                if not (selection is None or selection is False):
                    selection_mask = self.df.evaluate_selection_mask(selection, i1=i1, i2=i2, cache=True)  # TODO
                    # TODO: we probably want a way to avoid a to numpy conversion?
                    selection_mask = np.asarray(selection_mask)
                    references.append(selection_mask)
                    # some aggregators make a distiction between missing value and no value
                    # like nunique, they need to know if they should take the value into account or not
                    if hasattr(agg, 'set_selection_mask'):
                        agg.set_selection_mask(selection_mask)
                if agg_desc.expressions:
                    assert len(agg_desc.expressions) in [1, 2], "only length 1 or 2 supported for now"
                    for i, expression in enumerate(agg_desc.expressions):
                        block = block_map[agg_desc.expressions[i]]
                        dtype = self.dtypes[agg_desc.expressions[i]]
                        # we have data for the aggregator as well
                        if np.ma.isMaskedArray(block):
                            block, mask = block.data, np.ma.getmaskarray(block)
                            block = check_array(block, dtype)
                            agg.set_data(block, i)
                            references.extend([block])
                            if selection_mask is None:
                                selection_mask = ~mask
                            else:
                                selection_mask = selection_mask & ~mask
                            references.append(selection_mask)
                        else:
                            block = check_array(block, dtype)
                            agg.set_data(block, i)
                            references.extend([block])
                # we only have 1 data mask, since it's locally combined
                if selection_mask is not None:
                    agg.set_data_mask(selection_mask)
                    references.extend([selection_mask])
        grid.bin(all_aggregators, N)
        self.has_values = True

    def reduce(self, others):
        for agg_index, (agg_desc, selections, aggregation, selection_waslist) in enumerate(self.aggregations):
            for selection_index, selection in enumerate(selections):
                agg0 = aggregation[selection_index]
                aggs = [other.aggregations[agg_index][2][selection_index] for other in others]
                agg0.reduce(aggs)

    def get_result(self):
        results = []
        for agg_index, (agg_desc, selections, aggregation, selection_waslist) in enumerate(self.aggregations):
            grids = []
            for selection_index, selection in enumerate(selections):
                grid = agg_desc.get_result(aggregation[selection_index])
                grids.append(grid)
            result = np.asarray(grids) if selection_waslist else grids[0]
            if agg_desc.dtype_out != str:
                dtype_out = vaex.utils.to_native_dtype(agg_desc.dtype_out)
                dtype_out = vaex.array_types.to_numpy_type(dtype_out)
                result = result.view(dtype_out)
            result = result.copy()
            results.append(result)
        return results

    def get_values(self):
        values_outer = []
        for agg_index, (agg_desc, selections, aggregation, selection_waslist) in enumerate(self.aggregations):
            values = []
            for selection_index, selection in enumerate(selections):
                agg = aggregation[selection_index]
                values.append(np.asarray(agg))
            values_outer.append(values)
        return values_outer


    @classmethod
    def decode(cls, encoding, spec, df):
        # aggs = [vaex.agg._from_spec(agg_spec) for agg_spec in spec['aggregations']]
        aggs = encoding.decode_list('aggregation', spec['aggregations'])
        dtypes = encoding.decode_dict('dtype', spec['dtypes'])
        grid = encoding.decode('grid', spec['grid'])
        values = encoding.decode_list2('ndarray', spec['values']) if 'values' in spec else None
        # dtypes = {expr: _deserialize_type(type_spec) for expr, type_spec in spec['dtypes'].items()}
        for agg in aggs:
            agg._prepare_types(df)
        return cls(df, grid, aggs, dtypes, initial_values=values)

    def encode(self, encoding):
        # TODO: get rid of dtypes
        encoded = {'task': type(self).name,
                'grid': encoding.encode('grid', self.grid),
                'aggregations': encoding.encode_list("aggregation", self.aggregation_descriptions),
                'dtypes': encoding.encode_dict("dtype", self.dtypes)
                }
        if self.has_values:
            encoded['values'] = encoding.encode_list2('ndarray', self.get_values())
        return encoded



import configparser
import os
import io

import pyarrow as pa
import pyarrow.fs

from . import split_options, FileProxy, split_scheme
from .cache import CachedFile


def open_s3(path, mode='rb', fs_options={}):
    path, options = split_options(path, fs_options)
    use_cache = options.pop('cache', 'true' if mode == 'rb' else 'false') in [True, 'true', 'True', '1']
    file_system, path = parse(path, options)
    if use_cache:
        # we lazily open the file, if all is cached, we don't need to connect to s3
        if mode != "rb":
            raise ValueError(f'Cannot combine s3 with mode other than "rb", {mode} is not supported')
        def dup():
            return file_system.open_input_file(path)
        fp = lambda: FileProxy(file_system.open_input_file(path), path, dup)
        return CachedFile(fp, path)
    else:
        def arrow_open():
            if mode == "rb":
                return file_system.open_input_file(path)
            elif mode == "r":
                return io.TextIOWrapper(file_system.open_input_file(path))
            elif mode == "wb":
                # although a stream, will suffice in most cases
                return file_system.open_output_stream(path)
            elif mode == "w":
                return io.TextIOWrapper(file_system.open_output_stream(path))
            else:
                raise ValueError(f'Only mode=rb/bw/r/w are supported, not {mode}')
        return FileProxy(arrow_open(), path, dup=arrow_open)


def parse(path, options):
    path, options = split_options(path, options)
    scheme, _ = split_scheme(path)
    if scheme == 's3':
        # anon is for backwards compatibility
        options['anonymous'] = (options.pop('anon', None) in [True, 'true', 'True', '1']) or (options.pop('anonymous', None) in [True, 'true', 'True', '1'])
        if 'profile' in options:
            # TODO: ideally, Apache Arrow should take a profile argument
            profile = options.pop('profile')
            config = configparser.ConfigParser()
            config.read('~/.aws/credentials')
            options['access_key'] = config[profile]['aws_access_key_id']
            options['secret_key'] = config[profile]['aws_secret_access_key']
        if 'region' not in options:
            # we use this to get the default region
            file_system, path = pa.fs.FileSystem.from_uri(path)
            options['region'] = file_system.region
        file_system = pa.fs.S3FileSystem(**options)
    else:
        file_system = None
    return file_system, path



# -*- coding: utf-8 -*-
from __future__ import division, print_function
import numpy as np
import vaex
from .tasks import Task, TaskMapReduce
from .utils import _parse_f
import six


def _asfloat(a):
    if a.dtype.type == np.float64 and a.strides[0] == 8:
        return a
    else:
        return a.astype(np.float64, copy=False)

class TaskMapReduceLegacy(TaskMapReduce):
    def __init__(self, *args, **kwargs):
        kwargs = kwargs.copy()
        kwargs['ignore_filter'] = True
        TaskMapReduce.__init__(self, *args, **kwargs)

class TaskHistogram(Task):
    def __init__(self, df, subspace, expressions, size, limits, masked=False, weight=None):
        self.size = size
        self.limits = limits
        Task.__init__(self, df, expressions, name="histogram")
        self.subspace = subspace
        self.dtype = np.float64
        self.masked = masked
        self.weight = weight
        # self.grids = vaex.grids.Grids(self.df, self.df.executor.thread_pool, *expressions)
        # self.grids.ranges = limits
        # self.grids.grids["counts"] = vaex.grids.Grid(self.grids, size, self.dimension, None)
        shape1 = (self.size,) * self.dimension
        try:
            self.size[0]
            shape1 = tuple(self.size)
        except:
            pass
        shape = (self.subspace.executor.thread_pool.nthreads,) + shape1
        self.data = np.zeros(shape, dtype=self.dtype)
        self.ranges_flat = []
        self.minima = []
        self.maxima = []
        for limit in self.limits:
            self.ranges_flat.extend(limit)
            vmin, vmax = limit
            self.minima.append(vmin)
            self.maxima.append(vmax)
        if self.weight is not None:
            self.expressions_all.append(weight)
        # print self.ranges_flat

    def __repr__(self):
        name = self.__class__.__module__ + "." + self.__class__.__name__
        return "<%s(df=%r, expressions=%r, size=%r, limits=%r)> instance at 0x%x" % (name, self.df, self.expressions, self.size, self.limits, id(self))

    def map(self, thread_index, i1, i2, filter_mask, *blocks):
        class Info(object):
            pass
        info = Info()
        info.i1 = i1
        info.i2 = i2
        info.first = i1 == 0
        info.last = i2 == self.df.length_unfiltered()
        info.size = i2 - i1
        # print "bin", i1, i2, info.last
        # self.grids["counts"].bin_block(info, *blocks)
        # mask = self.df.mask
        data = self.data[thread_index]

        blocks = [_asfloat(block) for block in blocks]

        if self.masked or self.df.filtered:
            mask = self.df.evaluate_selection_mask("default" if self.masked else None, i1=i1, i2=i2, pre_filtered=False)
            blocks = [block[mask] for block in blocks]

        subblock_weight = None
        if len(blocks) == len(self.expressions) + 1:
            subblock_weight = blocks[-1]
            blocks = list(blocks[:-1])
        # print subblocks[0]
        # print subblocks[1]

        if self.dimension == 1:
            vaex.vaexfast.histogram1d(blocks[0], subblock_weight, data, *self.ranges_flat)
        elif self.dimension == 2:
            # if subblock_weight is None:
            # #print "speedup?"
            # histogram_numba(blocks[0], blocks[1], subblock_weight, data, *self.ranges_flat)
            # else:
            vaex.vaexfast.histogram2d(blocks[0], blocks[1], subblock_weight, data, *self.ranges_flat)
            # vaex.vaexfast.statisticNd([blocks[0], blocks[1]], subblock_weight, data, self.minima, self.maxima, 0)
        elif self.dimension == 3:
            vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], subblock_weight, data, *self.ranges_flat)
        else:
            blocks = list(blocks)  # histogramNd wants blocks to be a list
            vaex.vaexfast.histogramNd(blocks, subblock_weight, data, self.minima, self.maxima)

        return i1
        # return map(self._map, blocks)#[self.map(block) for block in blocks]

    def reduce(self, results):
        for i in range(1, self.subspace.executor.thread_pool.nthreads):
            self.data[0] += self.data[i]
        return self.data[0]
        # return self.data


class SubspaceGridded(object):
    def __init__(self, subspace_bounded, grid, vx=None, vy=None, vcounts=None):
        self.subspace_bounded = subspace_bounded
        self.grid = grid
        self.vx = vx
        self.vy = vy
        self.vcounts = vcounts

    def vector(self, weightx, weighty, size=32):
        counts = self.subspace_bounded.gridded_by_histogram(size=size)
        vx = self.subspace_bounded.gridded_by_histogram(size=size, weight=weightx)
        vy = self.subspace_bounded.gridded_by_histogram(size=size, weight=weighty)
        return SubspaceGridded(self.subspace_bounded, self.grid, vx=vx, vy=vy, vcounts=counts)

    def filter_gaussian(self, sigmas=1):
        import scipy.ndimage
        return SubspaceGridded(self.subspace_bounded, scipy.ndimage.filters.gaussian_filter(self.grid, sigmas))

    def clip_relative(self, v1, v2):
        vmin = self.grid.min()
        vmax = self.grid.max()
        width = vmax - vmin
        return SubspaceGridded(self.subspace_bounded, np.clip(self.grid, vmin + v1 * width, vmin + v2 * width))

    def volr(self, **kwargs):
        import vaex.notebook
        return vaex.notebook.volr(subspace_gridded=self, **kwargs)

    def plot(self, axes=None, **kwargs):
        self.subspace_bounded.subspace.plot(np.log1p(self.grid), limits=self.subspace_bounded.bounds, axes=axes, **kwargs)

    def mean_line(self, axis=0, **kwargs):
        from matplotlib import pylab
        assert axis in [0, 1]
        other_axis = 0 if axis == 1 else 1
        xmin, xmax = self.subspace_bounded.bounds[axis]
        ymin, ymax = self.subspace_bounded.bounds[other_axis]
        x = vaex.utils.linspace_centers(xmin, xmax, self.grid.shape[axis])
        y = vaex.utils.linspace_centers(ymin, ymax, self.grid.shape[other_axis])
        print(y)
        if axis == 0:
            counts = np.sum(self.grid, axis=axis)
            means = np.sum(self.grid * y[np.newaxis, :].T, axis=axis) / counts
        else:
            counts = np.sum(self.grid, axis=axis)
            means = np.sum(self.grid * y[:, np.newaxis].T, axis=axis) / counts
        if axis == 0:
            result = pylab.plot(x, means, **kwargs)
        else:
            result = pylab.plot(means, x, **kwargs)

        self.subspace_bounded.lim()
        return result, x, means

    def _repr_png_(self):
        from matplotlib import pylab
        fig, ax = pylab.subplots()
        self.plot(axes=ax, f=np.log1p)
        import vaex.utils
        if all([k is not None for k in [self.vx, self.vy, self.vcounts]]):
            N = self.vx.grid.shape[0]
            bounds = self.subspace_bounded.bounds
            print(bounds)
            positions = [vaex.utils.linspace_centers(bounds[i][0], bounds[i][1], N) for i in range(self.subspace_bounded.subspace.dimension)]
            print(positions)
            mask = self.vcounts.grid > 0
            vx = np.zeros_like(self.vx.grid)
            vy = np.zeros_like(self.vy.grid)
            vx[mask] = self.vx.grid[mask] / self.vcounts.grid[mask]
            vy[mask] = self.vy.grid[mask] / self.vcounts.grid[mask]
            # vx = self.vx.grid / self.vcounts.grid
            # vy = self.vy.grid / self.vcounts.grid
            x2d, y2d = np.meshgrid(positions[0], positions[1])
            ax.quiver(x2d[mask], y2d[mask], vx[mask], vy[mask])
            # print x2d
            # print y2d
            # print vx
            # print vy
            # ax.quiver(x2d, y2d, vx, vy)
        ax.title.set_text(r"$\log(1+counts)$")
        ax.set_xlabel(self.subspace_bounded.subspace.expressions[0])
        ax.set_ylabel(self.subspace_bounded.subspace.expressions[1])
        # pylab.savefig
        # from .io import StringIO
        from six import StringIO
        file_object = StringIO()
        fig.canvas.print_png(file_object)
        pylab.close(fig)
        return file_object.getvalue()

    def cube_png(self, f=np.log1p, colormap="afmhot", file="cube.png"):
        if self.grid.shape != ((128,) * 3):
            logger.error("only 128**3 cubes are supported")
            return None
        colormap_name = "afmhot"
        import matplotlib.cm
        colormap = matplotlib.cm.get_cmap(colormap_name)
        mapping = matplotlib.cm.ScalarMappable(cmap=colormap)
        # pixmap = QtGui.QPixmap(32*2, 32)
        data = np.zeros((128 * 8, 128 * 16, 4), dtype=np.uint8)

        # mi, ma = 1*10**self.mod1, self.data3d.max()*10**self.mod2
        grid = f(self.grid)
        vmin, vmax = grid.min(), grid.max()
        grid_normalized = (grid - vmin) / (vmax - vmin)
        # intensity_normalized = (np.log(self.data3d + 1.) - np.log(mi)) / (np.log(ma) - np.log(mi));
        import PIL.Image
        for y2d in range(8):
            for x2d in range(16):
                zindex = x2d + y2d * 16
                I = grid_normalized[zindex]
                rgba = mapping.to_rgba(I, bytes=True)  # .reshape(Nx, 4)
                # print rgba.shape
                subdata = data[y2d * 128:(y2d + 1) * 128, x2d * 128:(x2d + 1) * 128]
                for i in range(3):
                    subdata[:, :, i] = rgba[:, :, i]
                subdata[:, :, 3] = (grid_normalized[zindex] * 255).astype(np.uint8)  # * 0 + 255
                if 0:
                    filename = "cube%03d.png" % zindex
                    img = PIL.Image.frombuffer("RGB", (128, 128), subdata[:, :, 0:3] * 1)
                    print(("saving to", filename))
                    img.save(filename)
        img = PIL.Image.frombuffer("RGBA", (128 * 16, 128 * 8), data, 'raw')  # , "RGBA", 0, -1)
        # filename = "cube.png"
        # print "saving to", file
        img.save(file, "png")

        if 0:
            filename = "colormap.png"
            print(("saving to", filename))
            height, width = self.colormap_data.shape[:2]
            img = PIL.Image.frombuffer("RGB", (width, height), self.colormap_data)
            img.save(filename)


class SubspaceBounded(object):
    def __init__(self, subspace, bounds):
        self.subspace = subspace
        self.bounds = bounds

    def histogram(self, size=256, weight=None):
        return self.subspace.histogram(limits=self.bounds, size=size, weight=weight)

    def gridded(self, size=256, weight=None):
        return self.gridded_by_histogram(size=size, weight=weight)

    def gridded_by_histogram(self, size=256, weight=None):
        grid = self.histogram(size=size, weight=weight)
        return SubspaceGridded(self, grid)

    def lim(self):
        from matplotlib import pylab
        xmin, xmax = self.bounds[0]
        ymin, ymax = self.bounds[1]
        pylab.xlim(xmin, xmax)
        pylab.ylim(ymin, ymax)


class Subspaces(object):
    """
    :type: subspaces: list[Subspace]

    """

    def __init__(self, subspaces):
        self.subspaces = subspaces
        self.expressions = set()
        first_subspace = self.subspaces[0]
        self.delay = first_subspace.delay
        self.dimension = first_subspace.dimension
        self.df = self.subspaces[0].df
        for subspace in self.subspaces:
            assert subspace.df == self.subspaces[0].df
            assert subspace.delay == self.subspaces[0].delay
            assert subspace.dimension == self.subspaces[0].dimension, "subspace is of dimension %s, while first subspace if of dimension %s" % (subspace.dimension, self.subspaces[0].dimension)
            # assert subspace.sele== self.subspaces[0].delay
            self.expressions.update(subspace.expressions)
        self.expressions = list(self.expressions)
        self.subspace = self.df(*list(self.expressions), delay=self.delay, executor=first_subspace.executor)

    # def _repr_html_(self):

    def __len__(self):
        return len(self.subspaces)

    def names(self, seperator=" "):
        return [seperator.join(subspace.expressions) for subspace in self.subspaces]

    def expressions_list(self):
        return [subspace.expressions for subspace in self.subspaces]

    def selected(self):
        return Subspaces([subspace.selected() for subspace in self.subspaces])

    def _unpack(self, values):
        value_map = dict(zip(self.expressions, values))
        return [[value_map[ex] for ex in subspace.expressions] for subspace in self.subspaces]

    def _pack(self, values):
        value_map = {}
        for subspace_values, subspace in zip(values, self.subspaces):
            for value, expression in zip(subspace_values, subspace.expressions):
                if expression in value_map:
                    if isinstance(value, np.ndarray):
                        assert np.all(value_map[expression] == value), "inconsistency in subspaces, value for expression %r is %r in one case, and %r in the other" % (expression, value, value_map[expression])
                    else:
                        assert value_map[expression] == value, "inconsistency in subspaces, value for expression %r is %r in one case, and %r in the other" % (expression, value, value_map[expression])
                else:
                    value_map[expression] = value
        return [value_map[expression] for expression in self.expressions]

    def minmax(self):
        if self.delay:
            return self.subspace.minmax().then(self._unpack)
        else:
            return self._unpack(self.subspace.minmax())

    def limits_sigma(self, sigmas=3, square=False):
        if self.delay:
            return self.subspace.limits_sigma(sigmas=sigmas, square=square).then(self._unpack)
        else:
            return self._unpack(self.subspace.limits_sigma(sigmas=sigmas, square=square))

    def mutual_information(self, limits=None, size=256):
        if limits is not None:
            limits = self._pack(limits)

        def mutual_information(limits):
            return vaex.promise.listPromise([vaex.promise.Promise.fulfilled(subspace.mutual_information(subspace_limits, size=size)) for subspace_limits, subspace in zip(limits, self.subspaces)])
            # return histograms
        if limits is None:
            limits_promise = vaex.promise.Promise.fulfilled(self.subspace.minmax())
        else:
            limits_promise = vaex.promise.Promise.fulfilled(limits)
        limits_promise = limits_promise.then(self._unpack)
        promise = limits_promise.then(mutual_information)
        return promise if self.delay else promise.get()

    def mean(self):
        if self.delay:
            return self.subspace.mean().then(self._unpack)
        else:
            means = self.subspace.mean()
            return self._unpack(means)

    def var(self, means=None):
        # 'pack' means, and check if it makes sence
        if means is not None:
            means = self._pack(means)

        def var(means):
            return self.subspace.var(means=means)
        if self.delay:
            # if means is None:
            # return self.subspace.mean().then(var).then(self._unpack)
            # else:
            return var(means).then(self._unpack)
        else:
            # if means is None:
            # means = self.subspace.mean()
            # logger.debug("means: %r", means)
            return self._unpack(var(means=means))

    def correlation(self, means=None, vars=None):
        def var(means):
            return self.subspace.var(means=means)

        def correlation(means_and_vars):
            means, vars = means_and_vars
            means, vars = self._unpack(means), self._unpack(vars)
            # return self.subspace.correlation(means=means, vars=vars)
            return vaex.promise.listPromise([subspace.correlation(means=subspace_mean, vars=subspace_var) for subspace_mean, subspace_var, subspace in zip(means, vars, self.subspaces)])
        if means is not None:
            means = self._pack(means)
        if vars is not None:
            vars = self._pack(vars)
        if self.delay:
            if means is None:
                mean_promise = self.subspace.mean()
            else:
                mean_promise = vaex.promise.Promise.fulfilled(means)
            if vars is None:
                var_promise = mean_promise.then(var)
            else:
                var_promise = vaex.promise.Promise.fulfilled(vars)
            mean_and_var_calculated = vaex.promise.listPromise(mean_promise, var_promise)
            return mean_and_var_calculated.then(correlation)
        else:
            if means is None:
                means = self.subspace.mean()
            if vars is None:
                vars = self.subspace.var(means=means)
            means = self._unpack(means)
            vars = self._unpack(vars)
            return [subspace.correlation(means=subspace_mean, vars=subspace_var) for subspace_mean, subspace_var, subspace in zip(means, vars, self.subspaces)]
            # return correlation((means, vars))

    # def bounded_by(self, limits_list):
    # return SubspacesBounded(SubspaceBounded(subspace, limits) for subspace, limit in zip(self.subspaces, limits_list))


class Subspace(object):
    """A Subspace represent a subset of columns or expressions from a df.

    subspace are not instantiated directly, but by 'calling' the df like this:

    >>> subspace_xy = some_df("x", "y")
    >>> subspace_r = some_df("sqrt(x**2+y**2)")

    See `vaex.df.Dataset` for more documentation.

    """

    def __init__(self, df, expressions, executor, delay, masked=False):
        """

        :param Dataset df: the df the subspace refers to
        :param list[str] expressions: list of expressions that forms the subspace
        :param Executor executor: responsible for executing the tasks
        :param bool delay: return answers directly, or as a promise
        :param bool masked: work on the selection or not
        :return:
        """
        self.df = df
        self.expressions = expressions
        self.executor = executor
        self.delay = delay
        self.is_masked = masked

    def __repr__(self):
        name = self.__class__.__module__ + "." + self.__class__.__name__
        return "<%s(df=%r, expressions=%r, delay=%r, is_masked=%r)> instance at 0x%x" % (name, self.df, self.expressions, self.delay, self.is_masked, id(self))

    @property
    def dimension(self):
        return len(self.expressions)

    def get_selection(self):
        return self.df.get_selection("default") if self.is_masked else None

    def is_selected(self):
        return self.is_masked

    def selected(self):
        return self.__class__(self.df, expressions=self.expressions, executor=self.executor, delay=self.delay, masked=True)

    def delayhronous(self):
        return self.__class__(self.df, expressions=self.expressions, executor=self.executor, delay=True, masked=self.is_masked)

    def image_rgba_save(self, filename, data=None, rgba8=None, **kwargs):
        if rgba8 is not None:
            data = self.image_rgba_data(rgba8=rgba8, **kwargs)
        if data is None:
            data = self.image_rgba_data(**kwargs)
        with open(filename, "wb") as f:
            f.write(data)

    def image_rgba_notebook(self, data=None, rgba8=None, **kwargs):
        if rgba8 is not None:
            data = self.image_rgba_data(rgba8=rgba8, **kwargs)
        if data is None:
            data = self.image_rgba_data(**kwargs)
        from IPython.display import display, Image
        return Image(data=data)

    def image_rgba_data(self, rgba8=None, format="png", pil_draw=False, **kwargs):
        import PIL.Image
        import PIL.ImageDraw
        from six import StringIO
        if rgba8 is None:
            rgba8 = self.image_rgba(**kwargs)
        img = PIL.Image.frombuffer("RGBA", rgba8.shape[:2], rgba8, 'raw')  # , "RGBA", 0, -1)
        if pil_draw:
            draw = PIL.ImageDraw.Draw(img)
            pil_draw(draw)

        f = StringIO()
        img.save(f, format)
        return f.getvalue()

    def image_rgba_url(self, rgba8=None, **kwargs):
        if rgba8 is None:
            rgba8 = self.image_rgba(**kwargs)
        import PIL.Image
        img = PIL.Image.frombuffer("RGBA", rgba8.shape[:2], rgba8, 'raw')  # , "RGBA", 0, -1)
        from six import StringIO
        f = StringIO()
        img.save(f, "png")
        from base64 import b64encode
        imgurl = "data:image/png;base64," + b64encode(f.getvalue()) + ""
        return imgurl

    def normalize_grid(self, grid):
        grid = grid * 1  # copy
        mask = (grid > 0) & np.isfinite(grid)
        if grid.sum():
            grid -= grid[mask].min()
            grid /= grid[mask].max()
        else:
            grid[:] = 0
        return grid

    def limits(self, value, square=False):
        """TODO: doc + server side implementation"""
        if isinstance(value, six.string_types):
            import re
            match = re.match(r"(\d*)(\D*)", value)
            if match is None:
                raise ValueError("do not understand limit specifier %r, examples are 90%, 3sigma")
            else:
                value, type = match.groups()
                import ast
                value = ast.literal_eval(value)
                type = type.strip()
                if type in ["s", "sigma"]:
                    return self.limits_sigma(value)
                elif type in ["ss", "sigmasquare"]:
                    return self.limits_sigma(value, square=True)
                elif type in ["%", "percent"]:
                    return self.limits_percentage(value)
                elif type in ["%s", "%square", "percentsquare"]:
                    return self.limits_percentage(value, square=True)
        if value is None:
            return self.limits_percentage(square=square)
        else:
            return value

    def image_rgba(self, grid=None, size=256, limits=None, square=False, center=None, weight=None, weight_stat="mean", figsize=None,
                   aspect="auto", f=lambda x: x, axes=None, xlabel=None, ylabel=None,
                   group_by=None, group_limits=None, group_colors='jet', group_labels=None, group_count=10, cmap="afmhot",
                   vmin=None, vmax=None,
                   pre_blend=False, background_color="white", background_alpha=1., normalize=True, color=None):
        f = _parse_f(f)
        if grid is None:
            limits = self.limits(limits)
            if limits is None:
                limits = self.limits_sigma()
            if group_limits is None and group_by:
                group_limits = tuple(self.df(group_by).minmax()[0]) + (group_count,)
            if weight_stat == "mean" and weight is not None:
                grid = self.bin_mean(weight, limits=limits, size=size, group_limits=group_limits, group_by=group_by)
            else:
                grid = self.histogram(limits=limits, size=size, weight=weight, group_limits=group_limits, group_by=group_by)
            if grid is None:  # cancel occured
                return
        import matplotlib.cm
        background_color = np.array(matplotlib.colors.colorConverter.to_rgb(background_color))
        if group_by:
            gmin, gmax, group_count = group_limits
            if isinstance(group_colors, six.string_types):
                group_colors = matplotlib.cm.get_cmap(group_colors)
            if isinstance(group_colors, matplotlib.colors.Colormap):
                group_count = group_limits[2]
                colors = [group_colors(k / float(group_count - 1.)) for k in range(group_count)]
            else:
                colors = [matplotlib.colors.colorConverter.to_rgba(k) for k in group_colors]
            total = np.sum(grid, axis=0).T
            # grid /= total
            mask = total > 0
            alpha = total - total[mask].min()
            alpha[~mask] = 0
            alpha = total / alpha.max()
            rgba = grid.T.dot(colors)

            def _norm(data):
                mask = np.isfinite(data)
                data = data - data[mask].min()
                data /= data[mask].max()
                return data
            rgba[..., 3] = (f(alpha))
            # rgba[...,3] = 1
            rgba[total == 0, 3] = 0.
            mask = alpha > 0
            if 1:
                for i in range(3):
                    rgba[..., i] /= total
                    # rgba[...,i] /= rgba[...,0:3].max()
                    rgba[~mask, i] = background_color[i]
            rgba = (np.swapaxes(rgba, 0, 1))
        else:
            if color:
                color = np.array(matplotlib.colors.colorConverter.to_rgba(color))
                rgba = np.zeros(grid.shape + (4,))
                rgba[..., 0:4] = color
                data = f(grid)
                mask = (grid > 0) & np.isfinite(data)
                if vmin is None:
                    vmin = data[mask].min()
                if vmax is None:
                    vmax = data[mask].max()
                if mask.sum():
                    data -= vmin
                    data /= vmax
                    data[~mask] = 0
                else:
                    data[:] = 0
                rgba[..., 3] = data
            else:
                cmap = matplotlib.cm.get_cmap(cmap)
                data = f(grid)
                if normalize:
                    mask = (data > 0) & np.isfinite(data)
                    if vmin is None:
                        vmin = data[mask].min()
                    if vmax is None:
                        vmax = data[mask].max()
                    if mask.sum():
                        data -= vmin
                        data /= vmax
                    else:
                        data[:] = 0
                    data[~mask] = 0
                data = np.clip(data, 0, 1)
                rgba = cmap(data)
                if normalize:
                    rgba[~mask, 3] = 0
                rgba[..., 3] = 1  # data
            # rgba8 = np.swapaxes(rgba8, 0, 1)
        # white = np.ones_like(rgba[...,0:3])
        if pre_blend:
            # rgba[...,3] = background_alpha
            rgb = rgba[..., :3].T
            alpha = rgba[..., 3].T
            rgb[:] = rgb * alpha + background_color[:3].reshape(3, 1, 1) * (1 - alpha)
            alpha[:] = alpha + background_alpha * (1 - alpha)
        rgba = np.clip(rgba, 0, 1)
        rgba8 = (rgba * 255).astype(np.uint8)
        return rgba8

    def plot_vectors(self, expression_x, expression_y, limits, wx=None, wy=None, counts=None, size=32, axes=None, **kwargs):
        import pylab
        # refactor: should go to bin_means_xy
        if counts is None:
            counts = self.histogram(size=size, limits=limits)
        if wx is None:
            wx = self.histogram(size=size, weight=expression_x, limits=limits)
        if wy is None:
            wy = self.histogram(size=size, weight=expression_y, limits=limits)
        N = size
        positions = [vaex.utils.linspace_centers(limits[i][0], limits[i][1], N) for i in range(self.dimension)]
        # print(positions)
        mask = counts > 0
        vx = wx / counts
        vy = wy / counts
        vx[counts == 0] = 0
        vy[counts == 0] = 0
        # vx = self.vx.grid / self.vcounts.grid
        # vy = self.vy.grid / self.vcounts.grid
        x2d, y2d = np.meshgrid(positions[0], positions[1])
        if axes is None:
            axes = pylab.gca()
        axes.quiver(x2d[mask], y2d[mask], vx[mask], vy[mask], **kwargs)

    def plot(self, grid=None, size=256, limits=None, square=False, center=None, weight=None, weight_stat="mean", figsize=None,
             aspect="auto", f="identity", axes=None, xlabel=None, ylabel=None,
             group_by=None, group_limits=None, group_colors='jet', group_labels=None, group_count=None,
             vmin=None, vmax=None,
             cmap="afmhot",
             **kwargs):
        """Plot the subspace using sane defaults to get a quick look at the data.

        :param grid: A 2d numpy array with the counts, if None it will be calculated using limits provided and Subspace.histogram
        :param size: Passed to Subspace.histogram
        :param limits: Limits for the subspace in the form [[xmin, xmax], [ymin, ymax]], if None it will be calculated using Subspace.limits_sigma
        :param square: argument passed to Subspace.limits_sigma
        :param Executor executor: responsible for executing the tasks
        :param figsize: (x, y) tuple passed to pylab.figure for setting the figure size
        :param aspect: Passed to matplotlib's axes.set_aspect
        :param xlabel: String for label on x axis (may contain latex)
        :param ylabel: Same for y axis
        :param kwargs: extra argument passed to axes.imshow, useful for setting the colormap for instance, e.g. cmap='afmhot'
        :return: matplotlib.image.AxesImage

         """
        import pylab
        f = _parse_f(f)
        limits = self.limits(limits)
        if limits is None:
            limits = self.limits_sigma()
        # if grid is None:
        if group_limits is None and group_by:
            group_limits = tuple(self.df(group_by).minmax()[0]) + (group_count,)
        # grid = self.histogram(limits=limits, size=size, weight=weight, group_limits=group_limits, group_by=group_by)
        if figsize is not None:
            pylab.figure(num=None, figsize=figsize, dpi=80, facecolor='w', edgecolor='k')
        if axes is None:
            axes = pylab.gca()
        fig = pylab.gcf()
        # if xlabel:
        pylab.xlabel(xlabel or self.expressions[0])
        # if ylabel:
        pylab.ylabel(ylabel or self.expressions[1])
        # axes.set_aspect(aspect)
        rgba8 = self.image_rgba(grid=grid, size=size, limits=limits, square=square, center=center, weight=weight, weight_stat=weight_stat,
                                f=f, axes=axes,
                                group_by=group_by, group_limits=group_limits, group_colors=group_colors, group_count=group_count,
                                vmin=vmin, vmax=vmax,
                                cmap=cmap)
        import matplotlib
        if group_by:
            if isinstance(group_colors, six.string_types):
                group_colors = matplotlib.cm.get_cmap(group_colors)
            if isinstance(group_colors, matplotlib.colors.Colormap):
                group_count = group_limits[2]
                colors = [group_colors(k / float(group_count - 1.)) for k in range(group_count)]
            else:
                colors = [matplotlib.colors.colorConverter.to_rgba(k) for k in group_colors]
            colormap = matplotlib.colors.ListedColormap(colors)
            gmin, gmax, group_count = group_limits  # [:2]
            delta = (gmax - gmin) / (group_count - 1.)
            norm = matplotlib.colors.Normalize(gmin - delta / 2, gmax + delta / 2)
            sm = matplotlib.cm.ScalarMappable(norm, colormap)
            sm.set_array(1)  # make matplotlib happy (strange behavious)
            colorbar = fig.colorbar(sm)
            if group_labels:
                colorbar.set_ticks(np.arange(gmin, gmax + delta / 2, delta))
                colorbar.set_ticklabels(group_labels)
            else:
                colorbar.set_ticks(np.arange(gmin, gmax + delta / 2, delta))
                colorbar.set_ticklabels(map(lambda x: "%f" % x, np.arange(gmin, gmax + delta / 2, delta)))
            colorbar.ax.set_ylabel(group_by)
            # matplotlib.colorbar.ColorbarBase(axes, norm=norm, cmap=colormap)
            im = axes.imshow(rgba8, extent=np.array(limits).flatten(), origin="lower", aspect=aspect, **kwargs)
        else:
            norm = matplotlib.colors.Normalize(0, 23)
            sm = matplotlib.cm.ScalarMappable(norm, cmap)
            sm.set_array(1)  # make matplotlib happy (strange behavious)
            colorbar = fig.colorbar(sm)
            im = axes.imshow(rgba8, extent=np.array(limits).flatten(), origin="lower", aspect=aspect, **kwargs)
            colorbar = None
        return im, colorbar

    def plot1d(self, grid=None, size=64, limits=None, weight=None, figsize=None, f="identity", axes=None, xlabel=None, ylabel=None, **kwargs):
        """Plot the subspace using sane defaults to get a quick look at the data.

        :param grid: A 2d numpy array with the counts, if None it will be calculated using limits provided and Subspace.histogram
        :param size: Passed to Subspace.histogram
        :param limits: Limits for the subspace in the form [[xmin, xmax], [ymin, ymax]], if None it will be calculated using Subspace.limits_sigma
        :param figsize: (x, y) tuple passed to pylab.figure for setting the figure size
        :param xlabel: String for label on x axis (may contain latex)
        :param ylabel: Same for y axis
        :param kwargs: extra argument passed to ...,

         """
        import pylab
        f = _parse_f(f)
        limits = self.limits(limits)
        assert self.dimension == 1, "can only plot 1d, not %s" % self.dimension
        if limits is None:
            limits = self.limits_sigma()
        if grid is None:
            grid = self.histogram(limits=limits, size=size, weight=weight)
        if figsize is not None:
            pylab.figure(num=None, figsize=figsize, dpi=80, facecolor='w', edgecolor='k')
        if axes is None:
            axes = pylab.gca()
        # if xlabel:
        pylab.xlabel(xlabel or self.expressions[0])
        # if ylabel:
        # pylab.ylabel(ylabel or self.expressions[1])
        pylab.ylabel("counts" or ylabel)
        # axes.set_aspect(aspect)
        N = len(grid)
        xmin, xmax = limits[0]
        return pylab.plot(np.arange(N) / (N - 1.0) * (xmax - xmin) + xmin, f(grid,), drawstyle="steps", **kwargs)
        # pylab.ylim(-1, 6)

    def plot_histogram_bq(self, f="identity", size=64, limits=None, color="red", bq_cleanup=True):
        import vaex.ext.bqplot
        limits = self.limits(limits)
        plot = vaex.ext.bqplot.BqplotHistogram(self, color, size, limits)
        if not hasattr(self, "_bqplot"):
            self._bqplot = {}
            self._bqplot["cleanups"] = []
        else:
            if bq_cleanup:
                for cleanup in self._bqplot["cleanups"]:
                    cleanup()
            self._bqplot["cleanups"] = []

        def cleanup(callback=plot.callback):
            self.df.signal_selection_changed.disconnect(callback=callback)
        self._bqplot["cleanups"].append(cleanup)

        return plot

    def plot_bq(self, grid=None, size=256, limits=None, square=False, center=None, weight=None, figsize=None,
                aspect="auto", f="identity", fig=None, axes=None, xlabel=None, ylabel=None, title=None,
                group_by=None, group_limits=None, group_colors='jet', group_labels=None, group_count=None,
                cmap="afmhot", scales=None, tool_select=False, bq_cleanup=True,
                **kwargs):
        import vaex.ext.bqplot
        import bqplot.interacts
        import bqplot.pyplot as p
        import ipywidgets as widgets
        import bqplot as bq
        f = _parse_f(f)
        limits = self.limits(limits)
        import vaex.ext.bqplot
        vaex.ext.bqplot.patch()
        if not hasattr(self, "_bqplot"):
            self._bqplot = {}
            self._bqplot["cleanups"] = []
        else:
            if bq_cleanup:
                for cleanup in self._bqplot["cleanups"]:
                    cleanup()
            self._bqplot["cleanups"] = []
        if limits is None:
            limits = self.limits_sigma()
        # if fig is None:
        if scales is None:
            x_scale = bq.LinearScale(min=limits[0][0], max=limits[0][1])
            y_scale = bq.LinearScale(min=limits[1][0], max=limits[1][1])
            scales = {'x': x_scale, 'y': y_scale}
        else:
            x_scale = scales["x"]
            y_scale = scales["y"]
        if 1:
            fig = p.figure()  # actually, bqplot doesn't return it
            fig = p.current_figure()
            fig.fig_color = "black"  # TODO, take the color from the colormap
            fig.padding_y = 0
            # if we don't do this, bqplot may flip some axes... report this bug
            x = np.arange(10)
            y = x**2
            p.plot(x, y, scales=scales)
            # p.xlim(*limits[0])
            # p.ylim(*limits[1])
            # if grid is None:
        if group_limits is None and group_by:
            group_limits = tuple(self.df(group_by).minmax()[0]) + (group_count,)
        # fig = p.
        # if xlabel:
        fig.axes[0].label = xlabel or self.expressions[0]
        # if ylabel:
        fig.axes[1].label = ylabel or self.expressions[1]
        if title:
            fig.title = title
        # axes.set_aspect(aspect)
        rgba8 = self.image_rgba(grid=grid, size=size, limits=limits, square=square, center=center, weight=weight,
                                f=f, axes=axes,
                                group_by=group_by, group_limits=group_limits, group_colors=group_colors, group_count=group_count,
                                cmap=cmap)
        # x_scale = p._context["scales"]["x"]
        # y_scale = p._context["scales"]["y"]
        src = "http://localhost:8888/kernelspecs/python2/logo-64x64.png"
        import bqplot.marks
        im = vaex.ext.bqplot.Image(src=src, scales=scales, x=0, y=0, width=1, height=1)
        if 0:
            size = 20
            x_data = np.arange(size)
            line = bq.Lines(x=x_data, y=np.random.randn(size), scales={'x': x_scale, 'y': y_scale},
                            stroke_width=3, colors=['red'])

            ax_x = bq.Axis(scale=x_scale, tick_format='0.2f', grid_lines='solid')
            ax_y = bq.Axis(scale=y_scale, orientation='vertical', tick_format='0.2f', grid_lines='solid')
            panzoom = bq.PanZoom(scales={'x': [x_scale], 'y': [y_scale]})
            lasso = bqplot.interacts.LassoSelector()
            brush = bqplot.interacts.BrushSelector(x_scale=x_scale, y_scale=y_scale, color="green")
            fig = bq.Figure(marks=[line, im], axes=[ax_x, ax_y], min_width=100, min_height=100, interaction=panzoom)
        else:
            fig.marks = list(fig.marks) + [im]

        def make_image(executor, limits):
            # print "make image" * 100
            self.executor = executor
            if self.df.has_selection():
                sub = self.selected()
            else:
                sub = self
            return sub.image_rgba(limits=limits, size=size, f=f)
        progress = widgets.FloatProgress(value=0.0, min=0.0, max=1.0, step=0.01)
        updater = vaex.ext.bqplot.DebouncedThreadedUpdater(self, size, im, make_image, progress_widget=progress)

        def update_image():
            limits = [x_scale.min, x_scale.max], [y_scale.min, y_scale.max]
            # print limits
            # print "update...", limits
            # vxbq.debounced_threaded_update(self.df, im, make_image2, limits=limits)
            updater.update(limits)

        def update(*args):
            update_image()
        y_scale.observe(update, "min")
        y_scale.observe(update, "max")
        x_scale.observe(update, "min")
        x_scale.observe(update, "max")
        update_image()
        # fig = kwargs.pop('figure', p.current_figure())
        tools = []
        tool_actions = []
        panzoom = bq.PanZoom(scales={'x': [x_scale], 'y': [y_scale]})
        tool_actions_map = {u"m": panzoom}
        tool_actions.append(u"m")

        fig.interaction = panzoom
        if tool_select:
            brush = bqplot.interacts.BrushSelector(x_scale=x_scale, y_scale=y_scale, color="green")
            tool_actions_map["b"] = brush
            tool_actions.append("b")

            def update_selection(*args):
                def f():
                    if brush.selected:
                        (x1, y1), (x2, y2) = brush.selected
                        ex1, ex2 = self.expressions
                        mode = modes_names[modes_labels.index(button_selection_mode.value)]
                        self.df.select_rectangle(ex1, ex2, limits=[[x1, x2], [y1, y2]], mode=mode)
                    else:
                        self.df.select_nothing()
                updater.update_select(f)
            brush.observe(update_selection, "selected")
            # fig.interaction = brush
            # callback = self.df.signal_selection_changed.connect(lambda df: update_image())
            callback = self.df.signal_selection_changed.connect(lambda df: updater.update_direct_safe())

            def cleanup(callback=callback):
                self.df.signal_selection_changed.disconnect(callback=callback)
            self._bqplot["cleanups"].append(cleanup)

            button_select_nothing = widgets.Button(icon="fa-trash-o")

            def select_nothing(button):
                self.df.select_nothing()
            button_select_nothing.on_click(select_nothing)
            tools.append(button_select_nothing)
            modes_names = "replace and or xor subtract".split()
            modes_labels = "= & | ^ -".split()
            button_selection_mode = widgets.ToggleButtons(description='', options=modes_labels)
            tools.append(button_selection_mode)

        def change_interact(*args):
            # print "change", args
            fig.interaction = tool_actions_map[button_action.value]
        # tool_actions = ["m", "b"]
        # tool_actions = [("m", "m"), ("b", "b")]
        button_action = widgets.ToggleButtons(description='', options=tool_actions, icons=["fa-arrows", "fa-pencil-square-o"])
        button_action.observe(change_interact, "value")
        tools.insert(0, button_action)
        button_action.value = "m"  # tool_actions[-1]
        if len(tools) == 1:
            tools = []
        tools = widgets.HBox(tools)

        box_layout = widgets.Layout(display='flex',
                                    flex_flow='column',
                                    # border='solid',
                                    width='100%', height="100%")
        fig.fig_margin = {'bottom': 40, 'left': 60, 'right': 10, 'top': 40}
        # fig.min_height = 700
        # fig.min_width = 400
        fig.layout = box_layout
        return widgets.VBox([fig, progress, tools])

    def figlarge(self, size=(10, 10)):
        import pylab
        pylab.figure(num=None, figsize=size, dpi=80, facecolor='w', edgecolor='k')

    # def bounded(self):
    # return self.bounded_by_minmax()

    def bounded_by(self, limits):
        """Returns a bounded subspace (SubspaceBounded) with limits as given by limits

        :param limits: sequence of [(min, max), ..., (min, max)] values
        :rtype: SubspaceBounded
        """
        return SubspaceBounded(self, np.array(limits))

    def bounded_by_minmax(self):
        """Returns a bounded subspace (SubspaceBounded) with limits given by Subspace.minmax()

        :rtype: SubspaceBounded
        """
        bounds = self.minmax()
        return SubspaceBounded(self, bounds)

    bounded = bounded_by_minmax

    def bounded_by_sigmas(self, sigmas=3, square=False):
        """Returns a bounded subspace (SubspaceBounded) with limits given by Subspace.limits_sigma()

        :rtype: SubspaceBounded
        """
        bounds = self.limits_sigma(sigmas=sigmas, square=square)
        return SubspaceBounded(self, bounds)

    def minmax(self):
        """Return a sequence of [(min, max), ..., (min, max)] corresponding to each expression in this subspace ignoring NaN.
        """
        raise NotImplementedError

    def mean(self):
        """Return a sequence of [mean, ... , mean] corresponding to the mean of each expression in this subspace ignoring NaN.
        """
        raise NotImplementedError

    def var(self, means=None):
        """Return a sequence of [var, ... , var] corresponding to the variance of each expression in this subspace ignoring NaN.
        """
        raise NotImplementedError

    def sum(self):
        """Return a sequence of [sum, ... , sum] corresponding to the sum of values of each expression in this subspace ignoring NaN."""
        raise NotImplementedError

    def histogram(self, limits, size=256, weight=None):
        """Return a grid of shape (size, ..., size) corresponding to the dimensionality of this subspace containing the counts in each element

        The type of the grid of np.float64

        """
        raise NotImplementedError

    def limits_sigma(self, sigmas=3, square=False):
        raise NotImplementedError

    def row(self, index):
        return np.array([self.df.evaluate(expression, i1=index, i2=index + 1)[0] for expression in self.expressions])


class SubspaceLocal(Subspace):
    """Subclass of subspace which implemented methods that can be run locally.
    """

    def _toarray(self, list):
        return np.array(list)

    @property
    def pre(self):
        self.executor.pre

    @property
    def post(self):
        self.executor.post

    def _task(self, task, progressbar=False):
        """Helper function for returning tasks results, result when immediate is True, otherwise the task itself, which is a promise"""
        if self.delay:
            # should return a task or a promise nesting it
            return self.executor.schedule(task)
        else:
            import vaex.utils
            callback = None
            try:
                if progressbar == True:
                    def update(fraction):
                        bar.update(fraction)
                        return True
                    bar = vaex.utils.progressbar(task.name)
                    callback = self.executor.signal_progress.connect(update)
                elif progressbar:
                    callback = self.executor.signal_progress.connect(progressbar)
                self.executor.schedule(task)
                self.df.execute()
                result = task.get()
                if progressbar == True:
                    bar.finish()
                    sys.stdout.write('\n')
                return result
            finally:
                if callback:
                    self.executor.signal_progress.disconnect(callback)

    def minmax(self, progressbar=False):
        def min_max_reduce(minmax1, minmax2):
            if minmax1 is None:
                return minmax2
            if minmax2 is None:
                return minmax1
            result = []
            for d in range(self.dimension):
                min1, max1 = minmax1[d]
                min2, max2 = minmax2[d]
                result.append((min(min1, min2), max(max1, max2)))
            return result

        def min_max_map(thread_index, i1, i2, *blocks):
            if self.is_masked or self.df.filtered:
                mask = self.df.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2, pre_filtered=False)
                blocks = [block[mask] for block in blocks]
                is_empty = all(~mask)
                if is_empty:
                    return None
            # with lock:
            # print blocks
            # with lock:
            # print thread_index, i1, i2, blocks
            blocks = [_asfloat(block) for block in blocks]
            return [vaex.vaexfast.find_nan_min_max(block) for block in blocks]
            if 0:  # TODO: implement using statisticNd and benchmark
                minmaxes = np.zeros((len(blocks), 2), dtype=float)
                minmaxes[:, 0] = np.inf
                minmaxes[:, 1] = -np.inf
                for i, block in enumerate(blocks):
                    vaex.vaexfast.statisticNd([], block, minmaxes[i, :], [], [], 2)
                # minmaxes[~np.isfinite(minmaxes)] = np.nan
                return minmaxes
        task = TaskMapReduceLegacy(self.df, self.expressions, min_max_map, min_max_reduce, self._toarray, info=True, name="minmax")
        return self._task(task, progressbar=progressbar)

    def mean(self):
        return self._moment(1)

    def _moment(self, moment=1):
        def mean_reduce(means_and_counts1, means_and_counts2):
            means_and_counts = []
            for (mean1, count1), (mean2, count2) in zip(means_and_counts1, means_and_counts2):
                means_and_counts.append([np.nansum([mean1 * count1, mean2 * count2]) / (count1 + count2), count1 + count2])
            return means_and_counts

        def remove_counts(means_and_counts):
            return self._toarray(means_and_counts)[:, 0]

        def mean_map(thread_index, i1, i2, *blocks):
            if self.is_masked or self.df.filtered:
                mask = self.df.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2, pre_filtered=False)
                return [(np.nanmean(block[mask]**moment), np.count_nonzero(~np.isnan(block[mask]))) for block in blocks]
            else:
                return [(np.nanmean(block**moment), np.count_nonzero(~np.isnan(block))) for block in blocks]
        task = TaskMapReduceLegacy(self.df, self.expressions, mean_map, mean_reduce, remove_counts, info=True)
        return self._task(task)

    def var(self, means=None):
        # variances are linear, use the mean to reduce
        def vars_reduce(vars_and_counts1, vars_and_counts2):
            vars_and_counts = []
            for (var1, count1), (var2, count2) in zip(vars_and_counts1, vars_and_counts2):
                vars_and_counts.append([np.nansum([var1 * count1, var2 * count2]) / (count1 + count2), count1 + count2])
            return vars_and_counts

        def remove_counts(vars_and_counts):
            return self._toarray(vars_and_counts)[:, 0]
        if self.is_masked or self.df.filtered:
            def var_map(thread_index, i1, i2, *blocks):
                mask = self.df.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2, pre_filtered=False)
                if means is not None:
                    return [(np.nanmean((block[mask] - mean)**2), np.count_nonzero(~np.isnan(block[mask]))) for block, mean in zip(blocks, means)]
                else:
                    return [(np.nanmean(block[mask]**2), np.count_nonzero(~np.isnan(block[mask]))) for block in blocks]
            task = TaskMapReduceLegacy(self.df, self.expressions, var_map, vars_reduce, remove_counts, info=True)
        else:
            def var_map(*blocks):
                if means is not None:
                    return [(np.nanmean((block - mean)**2), np.count_nonzero(~np.isnan(block))) for block, mean in zip(blocks, means)]
                else:
                    return [(np.nanmean(block**2), np.count_nonzero(~np.isnan(block))) for block in blocks]
            task = TaskMapReduceLegacy(self.df, self.expressions, var_map, vars_reduce, remove_counts)
        return self._task(task)

    def correlation(self, means=None, vars=None):
        if self.dimension != 2:
            raise ValueError("correlation is only defined for 2d subspaces, not %dd" % self.dimension)

        def do_correlation(means, vars):
            meanx, meany = means
            sigmax, sigmay = vars[0]**0.5, vars[1]**0.5

            def remove_counts_and_normalize(covar_and_count):
                covar, counts = covar_and_count
                return covar / counts / (sigmax * sigmay)

            def covars_reduce(covar_and_count1, covar_and_count2):
                if covar_and_count1 is None:
                    return covar_and_count2
                if covar_and_count2 is None:
                    return covar_and_count1
                else:
                    covar1, count1 = covar_and_count1
                    covar2, count2 = covar_and_count2
                    return [np.nansum([covar1, covar2]), count1 + count2]

            mask = self.df.mask

            def covar_map(thread_index, i1, i2, *blocks):
                # return [(np.nanmean((block[mask[i1:i2]]-mean)**2), np.count_nonzero(~np.isnan(block[mask[i1:i2]]))) for block, mean in zip(blocks, means)]
                blockx, blocky = blocks
                if self.is_masked:
                    blockx, blocky = blockx[mask[i1:i2]], blocky[mask[i1:i2]]
                counts = np.count_nonzero(~(np.isnan(blockx) | np.isnan(blocky)))
                if counts == 0:
                    return None
                else:
                    return np.nansum((blockx - meanx) * (blocky - meany)), counts

            task = TaskMapReduceLegacy(self.df, self.expressions, covar_map, covars_reduce, remove_counts_and_normalize, info=True)
            return self._task(task)
        if means is None:
            if self.delay:
                means_wrapper = [None]

                def do_vars(means):
                    means_wrapper[0] = means
                    return self.var(means)

                def do_correlation_wrapper(vars):
                    return do_correlation(means_wrapper[0], vars)
                return self.mean().then(do_vars).then(do_correlation_wrapper)
            else:
                means = self.mean()
                vars = self.var(means=means)
                return do_correlation(means, vars)
        else:
            if vars is None:
                if self.delay:
                    def do_correlation_wrapper(vars):
                        return do_correlation(means, vars)
                    return self.vars(means=means).then(do_correlation_wrapper)
                else:
                    vars = self.var(means)
                    return do_correlation(means, vars)
            else:
                if means is None:
                    means = self.mean()
                if vars is None:
                    vars = self.var(means=means)
                return do_correlation(means, vars)

    def sum(self):
        def nansum(x): return np.nansum(x, dtype=np.float64)
        # TODO: we can speed up significantly using our own nansum, probably the same for var and mean
        nansum = vaex.vaexfast.nansum
        if self.is_masked or self.df.filtered:
            task = TaskMapReduceLegacy(self.df,
                                 self.expressions, lambda thread_index, i1, i2, *blocks: [nansum(block[self.df.evaluate_selection_mask("default" if self.is_masked else None, i1=i1, i2=i2, pre_filtered=False)])
                                                                                          for block in blocks],
                                 lambda a, b: np.array(a) + np.array(b), self._toarray, info=True)
        else:
            task = TaskMapReduceLegacy(self.df, self.expressions, lambda *blocks: [nansum(block) for block in blocks], lambda a, b: np.array(a) + np.array(b), self._toarray)
        return self._task(task)

    def histogram(self, limits, size=256, weight=None, progressbar=False, group_by=None, group_limits=None):
        expressions = self.expressions
        if group_by:
            expressions = list(expressions) + [group_by]
            limits = list(limits) + [group_limits[:2]]  # [[group_limits[0] - 0,5, group_limits[1]+0.5]]
            # assert group_limits[2] == 1
            size = (group_limits[2],) + (size,) * (len(expressions) - 1)
        task = TaskHistogram(self.df, self, expressions, size, limits, masked=self.is_masked, weight=weight)
        return self._task(task, progressbar=progressbar)

    def bin_mean(self, expression, limits, size=256, progressbar=False, group_by=None, group_limits=None):
        # todo, fix progressbar into two...
        counts = self.histogram(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits)
        weighted = self.histogram(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits,
                                  weight=expression)
        mean = weighted / counts
        mean[counts == 0] = np.nan
        return mean

    def bin_mean_cyclic(self, expression, max_value, limits, size=256, progressbar=False, group_by=None, group_limits=None):
        # todo, fix progressbar into two...
        meanx = self.bin_mean(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits,
                              expression="cos((%s)/%r*2*pi)" % (expression, max_value))
        meany = self.bin_mean(limits=limits, size=size, progressbar=progressbar, group_by=group_by, group_limits=group_limits,
                              expression="sin((%s)/%r*2*pi)" % (expression, max_value))
        angles = np.arctan2(meany, meanx)
        values = ((angles + 2 * np.pi) % (2 * np.pi)) / (2 * np.pi) * max_value
        length = np.sqrt(meanx**2 + meany**2)
        length[~np.isfinite(meanx)] = np.nan
        return values, length

    def mutual_information(self, limits=None, grid=None, size=256):
        if limits is None:
            limits_done = Task.fulfilled(self.minmax())
        else:
            limits_done = Task.fulfilled(limits)
        if grid is None:
            if limits is None:
                histogram_done = limits_done.then(lambda limits: self.histogram(limits, size=size))
            else:
                histogram_done = Task.fulfilled(self.histogram(limits, size=size))
        else:
            histogram_done = Task.fulfilled(grid)
        mutual_information_promise = histogram_done.then(vaex.kld.mutual_information)
        return mutual_information_promise if self.delay else mutual_information_promise.get()

    def limits_percentage(self, percentage=99.73, square=False):
        import scipy.ndimage
        limits = []
        for expr in self.expressions:
            subspace = self.df(expr)
            if self.is_selected():
                subspace = subspace.selected()
            limits_minmax = subspace.minmax()
            vmin, vmax = limits_minmax[0]
            size = 1024 * 16
            counts = subspace.histogram(size=size, limits=limits_minmax)
            cumcounts = np.concatenate([[0], np.cumsum(counts)])
            cumcounts /= cumcounts.max()
            # TODO: this is crude.. see the details!
            f = (1 - percentage / 100.) / 2
            x = np.linspace(vmin, vmax, size + 1)
            l = scipy.interp([f, 1 - f], cumcounts, x)
            limits.append(l)
        return limits

    def limits_sigma(self, sigmas=3, square=False):
        if self.delay:
            means_wrapper = [None]

            def do_vars(means):
                means_wrapper[0] = means
                return self.var(means)

            def do_limits(vars):
                stds = vars**0.5
                means = means_wrapper[0]
                if square:
                    stds = np.repeat(stds.mean(), len(stds))
                return np.array(list(zip(means - sigmas * stds, means + sigmas * stds)))
            return self.mean().then(do_vars).then(do_limits)
        else:
            means = self.mean()
            stds = self.var(means=means)**0.5
            if square:
                stds = np.repeat(stds.mean(), len(stds))
            return np.array(list(zip(means - sigmas * stds, means + sigmas * stds)))

    def _not_needed_current(self):
        index = self.df.get_current_row()

        def find(thread_index, i1, i2, *blocks):
            if (index >= i1) and (index < i2):
                return [block[index - i1] for block in blocks]
            else:
                return None
        task = TaskMapReduceLegacy(self.df, self.expressions, find, lambda a, b: a if b is None else b, info=True)
        return self._task(task)

    def nearest(self, point, metric=None):
        metric = metric or [1.] * len(point)

        def nearest_in_block(thread_index, i1, i2, *blocks):
            if self.is_masked:
                mask = self.df.evaluate_selection_mask("default", i1=i1, i2=i2, pre_filtered=False)
                if mask.sum() == 0:
                    return None
                blocks = [block[mask] for block in blocks]
            distance_squared = np.sum([(blocks[i] - point[i])**2. * metric[i] for i in range(self.dimension)], axis=0)
            min_index_global = min_index = np.argmin(distance_squared)
            if self.is_masked:  # we skipped some indices, so correct for that
                min_index_global = np.argmin((np.cumsum(mask) - 1 - min_index)**2)
            # with lock:
            # print i1, i2, min_index, distance_squared, [block[min_index] for block in blocks]
            return min_index_global.item() + i1, distance_squared[min_index].item()**0.5, [block[min_index].item() for block in blocks]

        def nearest_reduce(a, b):
            if a is None:
                return b
            if b is None:
                return a
            if a[1] < b[1]:
                return a
            else:
                return b
        if self.is_masked:
            pass
        task = TaskMapReduceLegacy(self.df,
                             self.expressions,
                             nearest_in_block,
                             nearest_reduce, info=True)
        return self._task(task)



import vaex.file

class ArrowOpener:
    @classmethod
    def quick_test(cls, path, *args, **kwargs):
        return vaex.file.ext(path) == '.arrow'

    @classmethod
    def can_open(cls, path, *args, **kwargs):
        return cls.quick_test(path, *args, **kwargs)

    @staticmethod
    def open(path, *args, **kwargs):
        from .dataset import open
        return open(path, *args, **kwargs)

class ParquetOpener:
    @classmethod
    def quick_test(cls, path, *args, **kwargs):
        return vaex.file.ext(path) == '.parquet'

    @classmethod
    def can_open(cls, path, *args, **kwargs):
        return cls.quick_test(path, *args, **kwargs)

    @staticmethod
    def open(path, *args, **kwargs):
        from .dataset import open_parquet
        return open_parquet(path, *args, **kwargs)



__author__ = 'breddels'
import pathlib
import logging
from glob import glob as local_glob
import os
import re
import sys
from urllib.parse import parse_qs

import vaex.file.cache


normal_open = open
logger = logging.getLogger("vaex.file")


class FileProxy:
    '''Wraps a file object, giving it a name a dup() method

    The dup is needed since a file is stateful, and needs to be duplicated in threads
    '''
    def __init__(self, file, name, dup):
        self.file = file
        self.name = name
        self.dup = dup

    def write(self, *args):
        return self.file.write(*args)

    def read(self, *args):
        return self.file.read(*args)

    def seek(self, *args):
        return self.file.seek(*args)

    def readinto(self, *args):
        return self.file.readinto(*args)

    def tell(self):
        return self.file.tell()

    def close(self):
        return self.file.close()

    def __enter__(self, *args):
        return self

    def __exit__(self, *args):
        self.file.close()


def is_file_object(file):
    return hasattr(file, 'read') and hasattr(file, 'seek')


def file_and_path(file, mode='r', fs_options={}):
    if is_file_object(file):
        return file, stringyfy(file)
    else:
        file = open(file, mode=mode, fs_options=fs_options)
        return file, stringyfy(file)


def stringyfy(path):
    if hasattr(path, 'name'):  # passed in a file
        path = path.name
    try:
        # Pathlib support
        path = path.__fspath__()
    except AttributeError:
        pass
    if isinstance(path, str):
        return path


def split_scheme(path):
    if '://' in path:
        schema, path = path[:path.index('://')], path[path.index('://')+3:]
    else:
        schema = None
    return schema, path


def memory_mappable(path):
    path = stringyfy(path)
    scheme, _ = split_scheme(path)
    return scheme is None


def split_options(path, fs_options={}):
    if isinstance(path, list):
        paths = []
        previous_options = None
        for path in path:
            path, options = split_options(path, fs_options)
            if previous_options is not None:
                if previous_options != options:
                    raise ValueError(f'Inconsistent set of fs_options given: {previous_options} {options}')
            else:
                previous_options = options
            paths.append(path)
        return paths, previous_options

    match = re.match(r'(.*?)\?((&?[^=&?]+=[^=&?]+)+)', path)
    if match:
        naked_path, query = match.groups()[:2]
    else:
        naked_path = path
        query = ''
    options = fs_options.copy()
    options.update({key: values[0] for key, values in parse_qs(query).items()})
    return naked_path, options


def open_google_cloud(path, mode, fs_options={}):
    from .gcs import open
    return vaex.file.gcs.open(path, mode, fs_options=fs_options)


def open_s3_arrow(path, mode, fs_options={}):
    from .arrow import open_s3
    return open_s3(path, mode, fs_options=fs_options)


def open_s3fs(path, mode, fs_options={}):
    from .s3 import open
    return open(path, mode, fs_options=fs_options)


scheme_opener = {
    None: normal_open,
    's3': open_s3_arrow,
    's3fs': open_s3fs,
    'gs': open_google_cloud
}


def open(path, mode='rb', fs_options={}):
    if is_file_object(path):
        return path
    path = stringyfy(path)
    scheme, _ = split_scheme(path)
    opener = scheme_opener.get(scheme)
    if not opener:
        raise ValueError(f'Do not know how to open {path}')
    if scheme is None:
        if fs_options:
            raise ValueError(f'fs_options not supported for local files. You passed: {repr(fs_options)}.')
        return opener(path, mode)
    if scheme == 's3':
        # fallback to s3fs for windows
        import pyarrow as pa
        try:
            return opener(path, mode, fs_options=fs_options)
        except pa.lib.ArrowNotImplementedError:
            opener = scheme_opener['s3fs']
    return opener(path, mode, fs_options=fs_options)


def open_for_arrow(path, mode='rb', fs_options={}, mmap=False):
    '''When the file will be passed to arrow, we want file object arrow likes.

    This might avoid peformance issues with GIL, or call overhead.
    '''
    import pyarrow as pa
    if is_file_object(path):
        return path
    path = stringyfy(path)
    scheme, _ = split_scheme(path)
    if scheme is None:
        if fs_options:
            raise ValueError(f'fs_options not supported for local files. You passed: {repr(fs_options)}.')
        if mmap:
            return pa.memory_map(path, mode)
        else:
            return pa.OSFile(path, mode)
    else:
        opener = scheme_opener.get(scheme)
        if not opener:
            raise ValueError(f'Do not know how to open {path}')
        return opener(path, mode, fs_options=fs_options).file


def dup(file):
    """Duplicate a file like object, s3 or cached file supported"""
    if isinstance(file, (vaex.file.cache.CachedFile, FileProxy)):
        return file.dup()
    else:
        return normal_open(file.name, file.mode)


def glob_s3(path, fs_options={}):
    from .s3 import glob
    return glob(path, fs_options=fs_options)


def glob_google_cloud(path, fs_options={}):
    from .gcs import glob
    return glob(path, fs_options=fs_options)


globber_map = {
    None: local_glob,
    's3': glob_s3,
    's3fs': glob_s3,
    'gs': glob_google_cloud
}


def glob(path, **kwargs):
    path = stringyfy(path)
    scheme, _ = split_scheme(path)
    globber = globber_map.get(scheme)
    if not globber:
        raise ValueError(f'Do not know how to glob {path}')
    return globber(path, **kwargs)


def ext(path):
    path = stringyfy(path)
    path, options = split_options(path)
    return os.path.splitext(path)[1]



# -*- coding: utf-8 -*-
from __future__ import division, print_function
import asyncio
import threading
import queue
import math
import multiprocessing
import sys
import vaex.utils
import logging
import traceback
import cProfile
import concurrent.futures
import time
import os

from .itertools import buffer

logger = logging.getLogger("vaex.multithreading")

thread_count_default = os.environ.get('VAEX_NUM_THREADS', multiprocessing.cpu_count())
thread_count_default = int(thread_count_default)
thread_count_default_io = os.environ.get('VAEX_NUM_THREADS_IO', multiprocessing.cpu_count() + 1)
thread_count_default_io = int(thread_count_default_io)
main_pool = None
main_io_pool = None


def get_main_pool():
    global main_pool
    if main_pool is None:
        main_pool = ThreadPoolIndex()
    return main_pool


def get_main_io_pool():
    global main_io_pool
    if main_io_pool is None:
        main_io_pool = concurrent.futures.ThreadPoolExecutor(max_workers=thread_count_default_io)
    return main_io_pool





class ThreadPoolIndex(concurrent.futures.ThreadPoolExecutor):
    """Thread pools that adds a thread index as first argument to the callback passed to map

    This is useful if you keep a piece of memory (like ndgrid) per array
    """

    def __init__(self, max_workers=None, *args, **kwargs):
        if max_workers is None:
            max_workers = thread_count_default
        super(ThreadPoolIndex, self).__init__(max_workers, *args, **kwargs)
        self.lock = threading.Lock()
        self.thread_indices = iter(range(1000000))  # enough threads until 2100?
        self.local = threading.local()
        self.nthreads = self._max_workers
        self._debug_sleep = 0

    async def map_async(self, callable, iterator, count, on_error=None, progress=None, cancel=None, unpack=False, **kwargs_extra):
        progress = progress or (lambda x: True)
        cancelled = False

        def wrapped(*args, **kwargs):
            if not cancelled:
                if self.nthreads == 1:
                    self.local.index = 0
                with self.lock:
                    if not hasattr(self.local, 'index'):
                        self.local.index = next(self.thread_indices)
                if unpack:
                    args = args[0]  # it's passed as a tuple.. not sure why
                if self._debug_sleep:
                    # print("SLEEP", self._debug_sleep)
                    time.sleep(self._debug_sleep)
                return callable(self.local.index, *args, **kwargs, **kwargs_extra)
        # convert to list so we can count
        time_last = time.time() - 100
        min_delta_t = 1. / 10  # max 10 per second
        if self.nthreads == 1:  # when using 1 thread, it makes debugging easier (better stacktrace)
            iterator = self._map_async(wrapped, iterator)
        else:
            loop = asyncio.get_event_loop()
            iterator = (loop.run_in_executor(self, lambda value=value: wrapped(value)) for value in iterator)

        total = 0
        for i, value in buffer(enumerate(iterator), self._max_workers + 3):
            value = await value
            if value != None:
                total += value
            progress_value = (total) / count
            time_now = time.time()
            if progress_value == 1 or (time_now - time_last) > min_delta_t:
                time_last = time_now
                if progress(progress_value) is False:
                    cancelled = True
                    cancel()
            yield value

    def _map_async(self, callable, iterator):
        for i in iterator:
            future = asyncio.Future()
            future.set_result(callable(i))
            yield future



"""Convert between arrow and vaex/numpy columns/arrays without doing memory copies.

We eventually like to get rid of this, and upstream all to pyarrow.
"""
import pyarrow
import pyarrow as pa
import numpy as np
import vaex.column


def ensure_not_chunked(arrow_array):
    if isinstance(arrow_array, pa.ChunkedArray):
        if len(arrow_array.chunks) == 0:
            return arrow_array.chunks[0]
        table = pa.Table.from_arrays([arrow_array], ["single"])
        table_concat = table.combine_chunks()
        column = table_concat.columns[0]
        assert column.num_chunks == 1
        arrow_array = column.chunk(0)
    return arrow_array


def arrow_array_from_numpy_array(array):  # TODO: -=> rename with ensure
    if isinstance(array, vaex.column.ColumnString):
        # we make sure it's ColumnStringArrow
        array = vaex.column._to_string_column(array)
        if len(array.bytes) < (2**31):  # arrow uses signed ints
            # if possible, downcast (parquet does not support large strings)
            array = array.to_arrow(pa.string())
        else:
            array = array.to_arrow()
    if isinstance(array, (pa.Array, pa.ChunkedArray)):
        return array
    dtype = array.dtype
    mask = None
    if np.ma.isMaskedArray(array):
        mask = array.mask
        # arrow 0.16 behaves weird in this case https://github.com/vaexio/vaex/pull/639
        if mask is np.False_:
            mask = None
        elif mask is np.True_:
            raise ValueError('not sure what pyarrow does with mask=True')
        array = array.data
    if dtype.kind == 'S':
        type = pyarrow.binary(dtype.itemsize)
        arrow_array = pyarrow.array(array, type, mask=mask)
    else:
        if not dtype.isnative:
            array = array.astype(dtype.newbyteorder('='))
        arrow_array = pyarrow.Array.from_pandas(array, mask=mask)
    return arrow_array

from vaex.dataframe import Column


def arrow_string_array_from_buffers(bytes, offsets, null_bitmap):
    if offsets.dtype == np.int32:
        type = pa.string()
    elif offsets.dtype == np.int64:
        type = pa.large_string()
    else:
        raise ValueError(f'Unsupported dtype {offsets.dtype} for string offsets')
    return _arrow_binary_array_from_buffers(bytes, offsets, null_bitmap, type)


def _arrow_binary_array_from_buffers(bytes, offsets, null_bitmap, type):
    length = len(offsets)-1
    offsets = pa.py_buffer(offsets)
    bytes = pa.py_buffer(bytes)
    if null_bitmap is not None:
        null_bitmap = pa.py_buffer(null_bitmap)
    return pa.Array.from_buffers(type, length, [null_bitmap, offsets, bytes])


def column_from_arrow_array(arrow_array):
    # TODO: we may be able to pass chunked arrays
    arrow_array = ensure_not_chunked(arrow_array)
    arrow_type = arrow_array.type
    buffers = arrow_array.buffers()
    if len(buffers) == 2:
        return numpy_array_from_arrow_array(arrow_array)
    elif len(buffers) == 3 and arrow_array.type in [pyarrow.string(), pyarrow.large_string()]:
        bitmap_buffer, offsets, string_bytes = arrow_array.buffers()
        if arrow_array.null_count == 0:
            null_bitmap = None  # we drop any null_bitmap when there are no null counts
        else:
            null_bitmap = np.frombuffer(bitmap_buffer, 'uint8', len(bitmap_buffer))
        if arrow_array.type == pyarrow.string():
            offsets = np.frombuffer(offsets, np.int32, len(offsets)//4)
        else:
            offsets = np.frombuffer(offsets, np.int64, len(offsets)//8)
        if string_bytes is None:
            string_bytes = np.array([], dtype='S1')
        else:
            string_bytes = np.frombuffer(string_bytes, 'S1', len(string_bytes))
        offset = arrow_array.offset
        column = vaex.column.ColumnStringArrow(offsets[offset:], string_bytes, len(arrow_array), null_bitmap=null_bitmap)
        return column
    else:
        raise TypeError('type unsupported: %r' % arrow_type)


def numpy_array_from_arrow_array(arrow_array):
    arrow_array = ensure_not_chunked(arrow_array)
    arrow_type = arrow_array.type
    buffers = arrow_array.buffers()
    assert len(buffers) == 2
    bitmap_buffer, data_buffer = buffers
    offset = arrow_array.offset
    if isinstance(arrow_type, type(pyarrow.binary(1))):  # todo, is there a better way to typecheck?
        # mimics python/pyarrow/array.pxi::Array::to_numpy
        assert len(buffers) == 2
        dtype = "S" + str(arrow_type.byte_width)
        # arrow seems to do padding, check if it is all ok
        expected_length = arrow_type.byte_width * len(arrow_array)
        actual_length = len(buffers[-1])
        if actual_length < expected_length:
            raise ValueError('buffer is smaller (%d) than expected (%d)' % (actual_length, expected_length))
        array = np.frombuffer(buffers[-1], dtype, len(arrow_array))# TODO: deal with offset ? [arrow_array.offset:arrow_array.offset + len(arrow_array)]
    else:
        dtype = arrow_array.type.to_pandas_dtype()
    if np.bool_ == dtype:
        # TODO: this will also be a copy, we probably want to support bitmasks as well
        bitmap = np.frombuffer(data_buffer, np.uint8, len(data_buffer))
        array = numpy_bool_from_arrow_bitmap(bitmap, len(arrow_array) + offset)[offset:]
    else:
        array = np.frombuffer(data_buffer, dtype, len(arrow_array) + offset)[offset:]

    if bitmap_buffer is not None:
        bitmap = np.frombuffer(bitmap_buffer, np.uint8, len(bitmap_buffer))
        mask = numpy_mask_from_arrow_mask(bitmap, len(arrow_array) + offset)[offset:]
        array = np.ma.MaskedArray(array, mask=mask)
    assert len(array) == len(arrow_array)
    return array


def numpy_mask_from_arrow_mask(bitmap, length):
    # arrow uses a bitmap https://github.com/apache/arrow/blob/master/format/Layout.md
    # we do have to change the ordering of the bits
    return 1-np.unpackbits(bitmap).reshape((len(bitmap),8))[:,::-1].reshape(-1)[:length]


def numpy_bool_from_arrow_bitmap(bitmap, length):
    # arrow uses a bitmap https://github.com/apache/arrow/blob/master/format/Layout.md
    # we do have to change the ordering of the bits
    return np.unpackbits(bitmap).reshape((len(bitmap),8))[:,::-1].reshape(-1)[:length].view(np.bool_)


def arrow_table_from_vaex_df(ds, column_names=None, selection=None, strings=True, virtual=False):
    """Implementation of Dataset.to_arrow_table"""
    names = []
    arrays = []
    for name, array in ds.to_items(column_names=column_names, selection=selection, strings=strings, virtual=virtual):
        names.append(name)
        arrays.append(arrow_array_from_numpy_array(array))
    return pyarrow.Table.from_arrays(arrays, names)


def vaex_df_from_arrow_table(table):
    from .dataset import DatasetArrow
    return DatasetArrow(table=table)


def trim_buffers(ar):
    # there are cases where memcopy are made, of modifications are mode (large_string_to_string)
    # in those cases, we don't want to work on the full array, and get rid of the offset if possible
    if ar.type == pa.string() or ar.type == pa.large_string():
        null_bitmap, offsets_buffer, bytes = ar.buffers()
        if ar.type == pa.string():
            offsets = np.frombuffer(offsets_buffer, np.int32, len(ar) + 1 + ar.offset)
        else:
            offsets = np.frombuffer(offsets_buffer, np.int64, len(ar) + 1 + ar.offset)
        # because it is difficult to slice bits
        new_offset = ar.offset % 8
        remove_offset = (ar.offset // 8) * 8
        first_offset = offsets[remove_offset]
        new_offsets = offsets[remove_offset:] - first_offset
        if null_bitmap:
            null_bitmap = null_bitmap.slice(ar.offset // 8)
        new_offsets_buffer = pa.py_buffer(new_offsets)
        bytes = bytes.slice(first_offset)
        ar = pa.Array.from_buffers(ar.type, len(ar), [null_bitmap, new_offsets_buffer, bytes], offset=new_offset)
    return ar


def large_string_to_string(ar):
    ar = trim_buffers(ar)
    offset = ar.offset
    null_bitmap, offsets_buffer, bytes = ar.buffers()
    offsets = np.frombuffer(offsets_buffer, np.int64, len(ar)+1 + ar.offset)
    if offsets[-1] > (2**31-1):
        raise ValueError('pa.large_string cannot be converted to pa.string')
    offsets = offsets.astype(np.int32)
    offsets_buffer = pa.py_buffer(offsets)
    return pa.Array.from_buffers(pa.string(), len(ar), [null_bitmap, offsets_buffer, bytes], offset=ar.offset)



__author__ = 'maartenbreddels'
import threading
import time
import queue
import logging

logger = logging.getLogger("vaex.ext.common")


class Job(object):
    def __init__(self, f, delay=0.0, **kwargs):
        self.cancelled = False
        self.f = f
        self.delay = delay
        self.kwargs = kwargs
        #self.thread = threading.Thread(target=self)
        self.result = None
        self.exception = False
        #self.finished = False

    def ____schedule(self):
        self.thread.start()

    def cancel(self):
        self.cancelled = True

    def __call__(self):
        time.sleep(self.delay)
        if not self.cancelled:
            try:
                self.result = self.f(self, **self.kwargs)
            except:
                self.exception = True
                raise



class SingleJobThread(threading.Thread):
    def __init__(self):
        self.job_queues = collections.defaultdict(queue.Queue)
        self.job_any = threading.Semaphore()
        self.start()

    def submit(self, job, queue_name):
        self.job_queues[queue_name].put(job)
        self.job_any.release()

    def run(self):
        try:
            logger.debug("vaex job thread is running")
            while True:
                self.job_any.acquire() # ok, there must be something to do...
                for key in self.job_queues.keys():
                    queue = self.job_queues[key]
                    if not queue.empty():
                        logging.debug("handling queue %s", key)
                        job = queue.get()
                        job()
        finally:
            logger.error("vaex job thread quits")




def buffer(i, n=1):
    values = []
    try:
        for _ in range(n-1):
            values.append(next(i))
        while True:
            values.append(next(i))
            yield values.pop(0)
    except StopIteration:
        pass
    yield from values


def pmap(f, i, pool):
    for item in i:
        def call(args):
            return f(*args)
        yield pool.submit(call, item)


def pwait(i):
    for item in i:
        yield item.result()


def consume(i):
    for item in i:
        pass
